import logging
import os
import numpy as np
import re
from ase.io import read
from ase import units as ase_units

from .metainfo import m_env
from nomad.units import ureg
from nomad.parsing import FairdiParser
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.common_dft import Run, BasisSetCellDependent, System, Method,\
    SingleConfigurationCalculation, ScfIteration, XCFunctionals, Eigenvalues, Workflow


re_float = r'[\d\.\-\+Ee]+'


class EigenvalueParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_eigenvalues(val_in):
            val = [v.split() for v in val_in.strip().split('\n')]
            kpoint = [0., 0., 0.] if val[0][0] == '#st' else [float(v.rstrip(',')) for v in val[0]]
            if len(val) == 1:
                return
            eigenvalues = np.array([[v[0], v[2], v[3]] for v in val[1:]], dtype=float)
            eigenvalues = np.transpose(eigenvalues)
            nspin = 2 if eigenvalues[0][1] == 1.0 else 1
            nbands = int(max(eigenvalues[0]))
            eigenvalues = eigenvalues[1:]
            eigenvalues.shape = (2, nbands, nspin)
            return kpoint, eigenvalues[0], eigenvalues[1]

        def str_to_fermi_energy(val_in):
            val = val_in.split()
            unit = ureg.eV if val[1].startswith('e') else ureg.hartree
            return float(val[0]) * unit

        self._quantities = [Quantity(
            'eigenvalues',
            r'(Eigenvalues \[[\s\S]+?)(?:\n\n|\Z)', sub_parser=TextParser(quantities=[
                Quantity(
                    'eigenvalues',
                    r'(?:(#st)\s*Spin\s*Eigenvalue\s*Occupation|#k =\s*\d+, k = \(([\-\d\.,\s]+)\))([\d\s\.\-updn]+)',
                    str_operation=str_to_eigenvalues, repeats=True, convert=False),
                Quantity(
                    'unit',
                    r'Eigenvalues \[(.*)\]', convert=False,
                    str_operation=lambda x: 'eV' if x == 'eV' else 'hartree'),
                Quantity(
                    'fermi_energy',
                    rf'Fermi energy =\s*({re_float} .*)',
                    convert=False, str_operation=str_to_fermi_energy)]))]


class InfoParser(EigenvalueParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):

        def str_to_energies(val_in):
            val = [v.split('=') for v in val_in.strip().split('\n')]
            unit = ureg.eV if '[eV]:' in val[0] else ureg.hartree
            return {v[0].strip(): float(v[1]) * unit for v in val if len(v) == 2}

        def str_to_forces(val_in):
            val = [v.split() for v in val_in.strip().split('\n')]
            unit = ureg.eV / ureg.angstrom if '[eV/A]' in val[0] else ureg.hartree / ureg.bohr
            return np.array([v[2:5] for v in val if len(v) == 5], dtype=float) * unit

        super().init_quantities()
        self._quantities += [
            Quantity(
                'brilloiun_zone_sampling',
                r'Brillouin zone sampling([\s\S]+?)\*+\n\n', sub_parser=TextParser(quantities=[
                    Quantity(
                        'grid_dimensions',
                        r'Dimensions of the k\-point grid\s*=(.*)', dtype=int),
                    Quantity(
                        'n_kpoints', r'Total number of k\-points\s*=(.*)', dtype=int),
                    Quantity(
                        'n_kpoints_reduced',
                        r'Number of symmetry-reduced k\-points\s*=(.*)', dtype=int),
                    Quantity(
                        'kpoints',
                        rf'\d+\s*({re_float})\s*({re_float})\s*({re_float})\s*({re_float})',
                        dtype=float, repeats=True)])),
            Quantity(
                'energies',
                r'Energy (\[[\s\S]+?)\n\n', str_operation=str_to_energies),
            Quantity(
                'total_magnetic_moment',
                rf'Total Magnetic Moment:\s*mz\s*=\s*({re_float})'),
            Quantity(
                'local_magnetic_moments',
                r'Ion\s*mz\s*([\w\s\.\-]+?)\n\n', convert=False,
                str_operation=lambda x: np.array(
                    [v.split()[2] for v in x.strip().split('\n')], dtype=float)),
            Quantity(
                'dipole',
                rf'Dipole:.*\[Debye\]\s*<x> =\s*\S+\s*({re_float})\s*'
                rf'<y> =\s*\S+\s*({re_float})\s*<z> =\s*\S+\s*({re_float})', dtype=float),
            Quantity(
                'forces',
                r'Forces on the ions (\[.*\]\s*)Ion\s*x\s*y\s*z\s*([\s\S]*)',
                str_operation=str_to_forces, convert=False)]


class ControlParser(TextParser):
    def __init__(self):
        super().__init__()
        self._constants = {
            'pi': np.pi, 'angstrom': 1. / ase_units.Bohr, 'ev': 1. / ase_units.Hartree,
            'yes': True, 'no': False, 't': True, 'f': False, 'true': True, 'false': False,
            'i': 1j}
        self._re_sqrt = re.compile(r'sqrt([\w\. ]+\Z)')
        self._info = None

    def init_quantities(self):
        def str_to_line(val_in):
            val = val_in.replace('"', '').replace("'", '').split('=', 1)
            return [v.strip().split('#')[0] for v in val]

        self._quantities = [
            Quantity(
                'line', r'(\w.*\s*=\s*.*)#?', str_operation=str_to_line, repeats=True)]

    @property
    def mainfile(self):
        return super().mainfile

    @mainfile.setter
    def mainfile(self, val):
        # TODO pylint complains when inheriting property
        self._results = None
        self._file_handler = None
        self._mainfile = os.path.abspath(val) if val is not None else val
        self._info = None
        self._keys_mapping = dict()

    def evaluate_value(self, value):
        # evaluate octopus parameterized variables, e.g. 2*angstrom
        # TODO implement all operations and parameters supported by Octopus

        if isinstance(value, list):
            return [self.evaluate_value(v) for v in value]

        if not isinstance(value, str):
            return value

        try:
            return float(value)
        except Exception:
            pass

        value = value.strip()
        if value == '':
            return 1

        val = self._constants.get(value.lower(), None)
        if val is not None:
            return val

        val = self._info.get(self._keys_mapping.get(value.lower(), None), None)
        if val is not None:
            return val

        open_p = value.rfind('(')
        if open_p > -1:
            n_groups = value.count('(')
            if n_groups != value.count(')'):
                return value
            for _ in range(n_groups):
                part = value[open_p + 1:]
                part = part[:part.find(')')]
                value = value.replace('(%s)' % part, str(self.evaluate_value(part)))
                open_p = value.rfind('(')
            return self.evaluate_value(value)

        sqrt = self._re_sqrt.match(value)
        if sqrt:
            val = self.evaluate_value(sqrt.group(1))
            return np.sqrt(val)

        vals = value.split('**')
        if len(vals) > 1:
            vals = [self.evaluate_value(v) for v in vals]
            val = vals[0]
            for v in reversed(vals[1:]):
                val = val ** v
            return val

        vals = value.split('*')
        if len(vals) > 1:
            vals = [self.evaluate_value(v) for v in vals]
            return np.product(vals)

        vals = value.split('/')
        if len(vals) > 1:
            vals = [self.evaluate_value(v) for v in vals]
            val = vals[0]
            for v in vals[1:]:
                val /= v
            return val

        if '+' in value:
            vals = value.split('+')
            vals = [self.evaluate_value(v) for v in vals]
            val = 0.0
            for v in vals:
                val += v
            return val

        if '-' in value:
            vals = value.split('-')
            vals = [self.evaluate_value(v) for v in vals]
            val = 0.0
            for v in vals:
                val -= v
            return val

        return value

    @property
    def info(self):
        if self._info is None:
            self._info = {v[0].strip(): v[1] for v in self.get('line', [])}
            self._info.update({v[0].strip(): v[1:] for v in self.get('block', [])})
            self._keys_mapping = {k.lower(): k for k in self._info.keys()}
            for key, val in self._info.items():
                try:
                    val = self.evaluate_value(val)
                except Exception:
                    continue
                self._info[key] = val
                self._keys_mapping[key.lower()] = key
        return self._info


class InpParser(ControlParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_block(val_in):
            val = [v.split('#')[0] for v in val_in.strip().split('\n')]
            val = [
                v.replace('"', '').replace("'", '').split('|')
                for v in val if v]
            val[0] = val[0][0]
            return val

        super().init_quantities()
        self._quantities += [
            Quantity(
                'block', r'%([\s\S]+?)%', repeats=True, str_operation=str_to_block)]

    def get_coordinates(self):
        val = self.info.get('Coordinates', self.info.get('ReducedCoordinates', []))

        symbols = []
        coordinates = []
        for v in val:
            symbols.append(v[0].strip())
            coordinates.append(v[1:])

        coordinates = np.array(coordinates, dtype=float)
        coordinates.shape = (len(symbols), 3)
        return symbols, coordinates


class LogParser(ControlParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_block(val_in):
            val = val_in.strip().split('\n')
            name = val[0].strip().replace('"', '').replace("'", '')
            val = [v.split('#')[0].split('=')[1].replace('"', '').replace("'", '').strip() for v in val[1:]]
            return [name, val]

        super().init_quantities()
        self._quantities += [
            Quantity(
                'block',
                r'Opened block([\s\S]+?)Closed block',
                repeats=True, str_operation=str_to_block)]

    def get_coordinates(self):
        symbols = []
        coordinates = []

        val = self.info.get('Coordinates', self.info.get('ReducedCoordinates', [[]]))[0]

        for v in val:
            if isinstance(v, str):
                symbols.append(v.strip())
            else:
                coordinates.append(v)

        coordinates = np.array(coordinates, dtype=float)
        coordinates.shape = (len(symbols), 3)

        return symbols, coordinates


class OutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_option(val_in):
            val = val_in.strip().split(':')
            return [val[0].strip(), ''.join(val[1:]).strip()]

        def str_to_cell(val_in):
            unit = ureg.angstrom if val_in.lower().startswith('a') else ureg.bohr
            val = [v.split() for v in val_in.strip().split('\n')[1:]]
            return np.array(val, dtype=float) * unit

        def str_to_spacing(val_in):
            unit = ureg.angstrom if val_in.startswith('A') else ureg.bohr
            return np.array(val_in.split()[1:4], dtype=float) * unit

        def str_to_energy(val_in):
            val = val_in.split()
            unit = ureg.eV if val[1].startswith('e') else ureg.hartree
            return float(val[0]) * unit

        def str_to_td_iteration(val_in):
            val = val_in.strip().split()
            return dict(
                iter=int(val[0]), time=float(val[1]), energy=float(val[2]),
                scfsteps=int(val[3]), elapsed_time=float(val[4]))

        iteration_quantities = [
            Quantity('energy_total', rf'etot\s*=\s*({re_float})'),
            # TODO scf_iteration eigenvalues are sometimes truncated and unusable
            Quantity(
                'fermi_level', rf'Fermi energy\s*=\s*({re_float} .*)',
                str_operation=str_to_energy, convert=False),
            Quantity(
                'time',
                r'Elapsed time for SCF step\s*\d+:\s*([\d\.]+)', unit='s', dtype=float)]

        self._quantities = [
            Quantity(
                'header',
                r'Running octopus([\s\S]+?)\*{10}', sub_parser=TextParser(quantities=[
                    Quantity(
                        'options',
                        r'\n *([\w ]+?\s*:\s*.*)', str_operation=str_to_option, repeats=True),
                ])),
            Quantity(
                'grid',
                r'\*\s*Grid\s*\*+\s*([\s\S]+?)\*{10}', sub_parser=TextParser(quantities=[
                    Quantity('boxshape', r'Type\s*=\s*(.*)'),
                    Quantity(
                        'npbc',
                        r'Octopus will treat the system as periodic in (\S+) dim', dtype=int),
                    Quantity(
                        'cell',
                        r'Lattice Vectors \[(.*)\]([-\d\s\.]+)',
                        str_operation=str_to_cell, convert=False),
                    Quantity(
                        'spacing',
                        r'Spacing \[(.*)\] = \(\s*(\S+), (\S+), (\S+)\s*\)',
                        str_operation=str_to_spacing, convert=False)])),
            Quantity(
                'theory_level',
                r'\*\s*Theory Level\s*\*+\s*([\s\S]+?)\*{10}', sub_parser=TextParser(quantities=[
                    Quantity('theory_level', r'\[TheoryLevel = (.+)\]', flatten=False),
                    Quantity('exchange', r'Exchange\s+(.*) \(', flatten=False),
                    Quantity('correlation', r'Correlation\s+(.*) \(', flatten=False)
                ])
            ),
            Quantity(
                'self_consistent',
                r'Info: Starting SCF iteration\.\s*([\s\S]+?)Info: SCF', repeats=True,
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        r'SCF CYCLE ITER #\s*(\d+\s*\*+[\s\S]+?)\*{10}', repeats=True,
                        sub_parser=TextParser(quantities=iteration_quantities))])),
            Quantity(
                'time_dependent',
                r'Time\-Dependent Simulation \*+([\s\S]+?)Info: Finished writing information',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        rf'\n *(\d+\s*{re_float}\s*{re_float}\s*\d+\s*{re_float}) *\n',
                        str_operation=str_to_td_iteration, repeats=True, convert=False)])
            ),
            Quantity(
                'x_octopus_info_scf_converged_iterations',
                r'SCF converged in\s*(\d+) iterations', dtype=int),
            Quantity(
                'minimization',
                r'(MINIMIZATION ITER #:\s*\d+\s*\++\s*Energy[\s\S]+?\+{10})',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'energy_total', rf'Energy\s*=\s*({re_float} .*)',
                        str_operation=str_to_energy, convert=False),
                    Quantity(
                        'number', r'ITER #:\s*(\d+)', dtype=int)]))
            # calculation results are not printed in outfile but in info
        ]

        self._header = None

    @property
    def header(self):
        return {k: v for k, v in self.get('header', {}).get('options', [])}


class OctopusParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/octopus', code_name='Octopus', code_homepage='https://octopus-code.org/',
            mainfile_contents_re=(r'\|0\) ~ \(0\) \|'))
        self._metainfo_env = m_env

        self.out_parser = OutParser()
        self.log_parser = LogParser()
        self.info_parser = InfoParser()
        self.inp_parser = InpParser()
        self.eigenvalue_parser = EigenvalueParser()

        self._info = None

        self._metainfo_keys_mapping = {
            'Total': 'energy_total', 'Free': 'energy_free', 'Ion-ion': 'x_octopus_info_energy_ion_ion',
            'Eigenvalues': 'energy_sum_eigenvalues', 'Hartree': 'energy_electrostatic',
            'Exchange': 'energy_X', 'Correlation': 'energy_C', 'vanderWaals': 'energy_van_der_Waals',
            '-TS': 'energy_correction_entropy', 'Kinetic': 'electronic_kinetic_energy',
            'SpinComponents': 'number_of_spin_channels', 'ExcessCharge': 'total_charge'}

        # TODO metainfo smearing_kind should be extended to cover all functions,
        # e.g. gaussian and spline are not the same
        self._smearing_functions = {
            1: 'empty', 2: 'fermi', 3: 'marzari-vanderbilt', 4: 'methfessel_paxton',
            5: 'gaussian', 'semiconducting': 'empty', 'spline_smearing': 'gaussian',
            'fermi_dirac': 'fermi', 'cold_smearing': 'marzari-vanderbilt',
            'methfessel_paxton': 'methfessel-paxton'}

        # TODO metainfo electronic_structure_method should be extended to cover all these
        # HF is not DFT but nomad treats it this way
        self._theory_levels = {
            1: 'DFT', 2: None, 3: 'DFT', 4: 'DFT', 5: None, 'hartree': 'DFT',
            'independent_particles': None, 'hartree_fock': 'DFT', 'dft': 'DFT',
            'classical': None}

        # List was generated from https://github.com/qsnake/octopus/blob/master/share/varinfo
        # In parser.log, XCFunctional is an integer given as a sum of exchange
        # and correlation. In stdout.txt, XCFunctional can be read as a string
        # but it is not always consistent with the mapping that we have below (see for
        # example Exchange which is Slater Exchange in stdout.txt). Therefore, in
        # resolving the nomad xc_functional name, we first try to get it from the string
        # output of stdout.txt and if it does not work, we get it from the integer value
        # in parser.log.
        # TODO resolve solely from the integer value
        self._xc_functionals = {
            'Exchange': ('lda_x', 1),
            'Wigner parametrization': ('lda_c_wigner', 2000),
            'Random Phase Approximation': ('lda_c_rpa', 3000),
            'Hedin & Lundqvist': ('lda_c_hl', 4000),
            'Gunnarson & Lundqvist': ('lda_c_gl', 5000),
            'Slater Xalpha': ('lda_c_xalpha', 6000),
            'Vosko, Wilk, & Nussair': ('lda_c_vwn', 7000),
            'Vosko, Wilk, & Nussair (RPA)': ('lda_c_vwn_rpa', 8000),
            'Perdew & Zunger': ('lda_c_pz', 9000),
            'Perdew & Zunger (Modified)': ('lda_c_pz_mod', 10000),
            'Ortiz & Ballone (PZ)': ('lda_c_ob_pz', 11000),
            'Perdew & Wang': ('lda_c_pw', 12000),
            'Perdew & Wang (Modified)': ('lda_c_pw_mod', 13000),
            'Ortiz & Ballone (PW)': ('lda_c_ob_pw', 14000),
            'Attacalite et al': ('lda_c_2d_amgb', 15000),
            'Pittalis, Rasanen & Marques correlation in 2D': ('lda_c_2d_prm', 16000),
            'von Barth & Hedin': ('lda_c_vbh', 17000),
            'Casula, Sorella, and Senatore 1D correlation': ('lda_c_1d_csc', 18000),
            'Exchange in 2D': ('lda_x_2d', 19),
            'Teter 93 parametrization': ('lda_xc_teter93', 20000),
            'Exchange in 1D': ('lda_x_1d', 21),
            'Modified LSD (version 1) of Proynov and Salahub': ('lda_c_ml1', 22000),
            'Modified LSD (version 2) of Proynov and Salahub': ('lda_c_ml2', 23000),
            'Gombas parametrization': ('lda_c_gombas', 24000),
            'Thomas-Fermi kinetic energy functional': ('lda_k_tf', 50),
            'Lee and Parr Gaussian ansatz': ('lda_k_lp', 51),
            'Perdew, Burke & Ernzerhof exchange': ('gga_x_pbe', 101),
            'Perdew, Burke & Ernzerhof exchange (revised)': ('gga_x_pbe_r', 102),
            'Becke 86 Xalfa,beta,gamma': ('gga_x_2d_b86', 128),
            'Herman et al original GGA': ('gga_x_herman', 104),
            'Becke 86 Xalfa,beta,gamma (with mod. grad. correction)': ('gga_x_b86_mgc', 105),
            'Becke 88': ('gga_x_b88', 106),
            'Gill 96': ('gga_x_g96', 107),
            'Perdew & Wang 86': ('gga_x_pw86', 108),
            'Perdew & Wang 91': ('gga_c_pw91', 134000),
            'Handy & Cohen OPTX 01': ('gga_x_optx', 110),
            'dePristo & Kress 87 (version R1)': ('gga_x_dk87_r1', 111),
            'dePristo & Kress 87 (version R2)': ('gga_x_dk87_r2', 112),
            'Lacks & Gordon 93': ('gga_x_lg93', 113),
            'Filatov & Thiel 97 (version A)': ('gga_x_ft97_a', 114),
            'Filatov & Thiel 97 (version B)': ('gga_x_ft97_b', 115),
            'Perdew, Burke & Ernzerhof exchange (solids)': ('gga_x_pbe_sol', 116),
            'Hammer, Hansen & Norskov (PBE-like)': ('gga_x_rpbe', 117),
            'Wu & Cohen': ('gga_x_wc', 118),
            'Modified form of PW91 by Adamo & Barone': ('gga_x_mpw91', 119),
            'Armiento & Mattsson 05 exchange': ('gga_x_am05', 120),
            'Madsen (PBE-like)': ('gga_x_pbea', 121),
            'Adamo & Barone modification to PBE': ('gga_x_mpbe', 122),
            'xPBE reparametrization by Xu & Goddard': ('gga_c_xpbe', 136000),
            'Becke 86 MGC for 2D systems': ('gga_x_2d_b86_mgc', 124),
            'Bayesian best fit for the enhancement factor': ('gga_x_bayesian', 125),
            'JSJR reparametrization by Pedroza, Silva & Capelle': ('gga_x_pbe_jsjr', 126),
            'Becke 88 in 2D': ('gga_x_2d_b88', 127),
            'Perdew, Burke & Ernzerhof exchange in 2D': ('gga_x_2d_pbe', 129),
            'Perdew, Burke & Ernzerhof correlation': ('gga_c_pbe', 130000),
            'Lee, Yang & Parr': ('gga_c_lyp', 131000),
            'Perdew 86': ('gga_c_p86', 132000),
            'Perdew, Burke & Ernzerhof correlation SOL': ('gga_c_pbe_sol', 133000),
            'Armiento & Mattsson 05 correlation': ('gga_c_am05', 135000),
            'Langreth and Mehl correlation': ('gga_c_lm', 137000),
            'JRGX reparametrization by Pedroza, Silva & Capelle': ('gga_c_pbe_jrgx', 138000),
            'Becke 88 reoptimized to be used with vdW functional of Dion et al': ('gga_x_optb88_vdw', 139),
            'PBE reparametrization for vdW': ('gga_x_optpbe_vdw', 141),
            'Regularized PBE': ('gga_c_rge2', 143000),
            'refitted Perdew & Wang 86': ('gga_x_rpw86', 144),
            'Keal and Tozer version 1': ('gga_x_kt1', 145),
            'Keal and Tozer version 2': ('gga_xc_kt2', 146000),
            'Wilson & Levy': ('gga_c_wl', 147000),
            'Wilson & Ivanov': ('gga_c_wi', 148000),
            'van Leeuwen & Baerends': ('gga_x_lb', 160),
            'HCTH functional fitted to 93 molecules': ('gga_xc_hcth_93', 161000),
            'HCTH functional fitted to 120 molecules': ('gga_xc_hcth_120', 162000),
            'HCTH functional fitted to 147 molecules': ('gga_xc_hcth_407', 164000),
            'Empirical functionals from Adamson, Gill, and Pople': ('gga_xc_edf1', 165000),
            'XLYP functional': ('gga_xc_xlyp', 166000),
            'Becke 97': ('hyb_gga_xc_b97', 407000),
            'Becke 97-1': ('hyb_gga_xc_b97_1', 408000),
            'Becke 97-2': ('hyb_gga_xc_b97_2', 410000),
            'Grimme functional to be used with C6 vdW term': ('gga_xc_b97_d', 170000),
            'Boese-Martin for Kinetics': ('hyb_gga_xc_b97_k', 413000),
            'Becke 97-3': ('hyb_gga_xc_b97_3', 414000),
            'Functionals fitted for water': ('gga_xc_pbelyp1w', 175000),
            'Schmider-Becke 98 parameterization 1a': ('hyb_gga_xc_sb98_1a', 420000),
            'Schmider-Becke 98 parameterization 1b': ('hyb_gga_xc_sb98_1b', 421000),
            'Schmider-Becke 98 parameterization 1c': ('hyb_gga_xc_sb98_1c', 422000),
            'Schmider-Becke 98 parameterization 2a': ('hyb_gga_xc_sb98_2a', 423000),
            'Schmider-Becke 98 parameterization 2b': ('hyb_gga_xc_sb98_2b', 424000),
            'Schmider-Becke 98 parameterization 2c': ('hyb_gga_xc_sb98_2c', 425000),
            'van Leeuwen & Baerends modified': ('gga_x_lbm', 182),
            'von Weiszaecker correction to Thomas-Fermi': ('gga_k_vw', 500),
            'Second-order gradient expansion (l = 1/9)': ('gga_k_ge2', 501),
            'TF-lambda-vW form by Golden (l = 13/45)': ('gga_k_golden', 502),
            'TF-lambda-vW form by Yonei and Tomishima (l = 1/5)': ('gga_k_yt65', 503),
            'TF-lambda-vW form by Baltin (l = 5/9)': ('gga_k_baltin', 504),
            'TF-lambda-vW form by Lieb (l = 0.185909191)': ('gga_k_lieb', 505),
            'gamma-TFvW form by Acharya et al [g = 1 - 1.412/N^(1/3)]': ('gga_k_absr1', 506),
            'gamma-TFvW form by Acharya et al [g = 1 - 1.332/N^(1/3)]': ('gga_k_absr2', 507),
            'gamma-TFvW form by Gázquez and Robles': ('gga_k_gr', 508),
            'gamma-TFvW form by Ludeña': ('gga_k_ludena', 509),
            'gamma-TFvW form by Ghosh and Parr': ('gga_k_gp85', 510),
            'Pearson': ('gga_k_pearson', 511),
            'Ou-Yang and Levy v.1': ('gga_k_ol1', 512),
            'Ou-Yang and Levy v.2': ('gga_k_ol2', 513),
            'Fuentealba & Reyes (B88 version)': ('gga_k_fr_b88', 514),
            'Fuentealba & Reyes (PW86 version)': ('gga_k_fr_pw86', 515),
            'The original hybrid proposed by Becke': ('hyb_gga_xc_b3pw91', 401000),
            'The (in)famous B3LYP': ('hyb_gga_xc_b3lyp', 402000),
            'Perdew 86 hybrid similar to B3PW91': ('hyb_gga_xc_b3p86', 403000),
            'hybrid using the optx functional': ('hyb_gga_xc_o3lyp', 404000),
            'mixture of mPW91 and PW91 optimized for kinetics': ('hyb_gga_xc_mpw1k', 405000),
            'aka PBE0 or PBE1PBE': ('hyb_gga_xc_pbeh', 406000),
            'maybe the best hybrid': ('hyb_gga_xc_x3lyp', 411000),
            'Becke 1-parameter mixture of WC and PBE': ('hyb_gga_xc_b1wc', 412000),
            'mixture with the mPW functional': ('hyb_gga_xc_mpw3pw', 415000),
            'Becke 1-parameter mixture of B88 and LYP': ('hyb_gga_xc_b1lyp', 416000),
            'Becke 1-parameter mixture of B88 and PW91': ('hyb_gga_xc_b1pw91', 417000),
            'Becke 1-parameter mixture of mPW91 and PW91': ('hyb_gga_xc_mpw1pw', 418000),
            'mixture of mPW and LYP': ('hyb_gga_xc_mpw3lyp', 419000),
            'Local tau approximation of Ernzerhof & Scuseria': ('mgga_x_lta', 201),
            'Perdew, Tao, Staroverov & Scuseria exchange': ('mgga_x_tpss', 202),
            'Zhao, Truhlar exchange': ('mgga_x_m06l', 203),
            'GVT4 from Van Voorhis and Scuseria (exchange part)': ('mgga_x_gvt4', 204),
            'tau-HCTH from Boese and Handy': ('mgga_x_tau_hcth', 205),
            'Becke-Roussel 89': ('mgga_x_br89', 206),
            'Becke & Johnson correction to Becke-Roussel 89': ('mgga_x_bj06', 207),
            'Tran & Blaha correction to Becke & Johnson': ('mgga_x_tb09', 208),
            'Rasanen, Pittalis, and Proetto correction to Becke & Johnson': ('mgga_x_rpp09', 209),
            'Pittalis, Rasanen, Helbig, Gross Exchange Functional': ('mgga_x_2d_prhg07', 210),
            'PRGH07 with PRP10 correction': ('mgga_x_2d_prhg07_prp10', 211),
            'Perdew, Tao, Staroverov & Scuseria correlation': ('mgga_c_tpss', 231000),
            'VSxc from Van Voorhis and Scuseria (correlation part)': ('mgga_c_vsxc', 232000),
            'Orestes, Marcasso & Capelle': ('lca_omc', 301),
            'Lee, Colwell & Handy': ('lca_lch', 302),
            'OEP: Exact exchange': ('oep_x', 901)}

        self._units_mapping = dict(ev=ureg.eV, hartree=ureg.hartree, angstrom=ureg.angstrom, bohr=ureg.bohr)

    def init_parser(self, filepath, logger):
        self.out_parser.mainfile = filepath
        self.out_parser.logger = logger

        self.info_parser.mainfile = os.path.join(self.maindir, 'static/info')
        self.info_parser.logger = logger
        self.log_parser.mainfile = os.path.join(self.maindir, 'exec/parser.log')
        self.log_parser.logger = logger
        self.inp_parser.mainfile = os.path.join(self.maindir, 'inp')
        self.inp_parser.logger = logger
        self.eigenvalue_parser.mainfile = os.path.join(self.maindir, 'static/eigenvalues')
        self.eigenvalue_parser.logger = logger
        self._info = None

    @property
    def info(self):
        if self._info is None:
            self._info = dict()
            self._info.update(self.inp_parser.info)
            # the variables read from log are more reliable as these are converted
            # to proper octopus values, i.e. we do not worry about compatibility issue
            self._info.update(self.log_parser.info)
            # add units for energy, length
            # TODO not sure what defaults are for units and expected values for units(input)
            # and their meaning
            energyunit = 'hartree'
            lengthunit = 'bohr'
            units = self._info.get('Units', self._info.get('UnitsInput', 0))
            if isinstance(units, str):
                units = units.lower()
                energyunit = 'eV' if 'ev' in units else energyunit
                lengthunit = 'angstrom' if 'angs' in units else lengthunit
            elif isinstance(units, int):
                energyunit = 'eV' if units == 1 else energyunit
                lengthunit = 'angstrom' if units == 1 else lengthunit
            elif units is None:
                cell = self.out_parser.get('grid', {}).get('cell')
                if cell is not None:
                    lengthunit = cell.units
                    energyunit = 'eV' if lengthunit == 'angstrom' else energyunit
            self._info.update(dict(energyunit=energyunit, lengthunit=lengthunit))
        return self._info

    def parse_scc(self):
        sec_run = self.archive.section_run[-1]

        # self-consistent
        # SCF energies are printed only in output
        for scf in self.out_parser.get('self_consistent', []):
            sec_scc = sec_run.m_create(SingleConfigurationCalculation)
            for iteration in scf.get('iteration', []):
                sec_scf = sec_scc.m_create(ScfIteration)
                fermi_level = iteration.get('fermi_level')
                unit = None
                if fermi_level is not None:
                    unit = fermi_level.units
                    fermi_level = [fermi_level.magnitude] * self.info.get('SpinComponents', 1)
                    sec_scf.energy_reference_fermi_iteration = fermi_level * unit
                energy_total = iteration.get('energy_total')
                if energy_total is not None:
                    unit = self._units_mapping.get(self.info.get('energyunit', 'hartree').lower()) if unit is None else unit
                    sec_scf.energy_total_scf_iteration = energy_total * unit
                time = iteration.get('time')
                if time is not None:
                    sec_scf.time_scf_iteration = time

        # time-dependent
        # each td iteration is an scc
        for td in self.out_parser.get('time_dependent', []):
            for iteration in td.get('iteration', []):
                sec_scc = sec_run.m_create(SingleConfigurationCalculation)
                energy = iteration.get('energy', None)
                if energy is not None:
                    sec_scc.energy_total = energy * self._units_mapping.get(self.info.get('energyunit', 'hartree').lower())

        # TODO add other calculation types

        # TODO bandstructures

        # TODO dipoles

        # TODO add volumetric data

        # TODO test geometry optimizations, cannot find an exaple
        for minimization in self.out_parser.get('minimization', []):
            sec_scc = sec_run.m_create(SingleConfigurationCalculation)
            energy_total = minimization.get('energy_total')
            if energy_total is not None:
                sec_scc.energy_total = energy_total
            number = minimization.get('number')
            cell = sec_run.section_system[-1].simulation_cell
            pbc = sec_run.section_system[-1].configuration_periodic_dimensions
            if number is not None:
                path = os.path.join(self.out_parser.maindir, 'geom/go.%04d.xyz' % number)
                try:
                    atoms = read(path, format='xyz')
                except Exception:
                    continue
                sec_system = sec_run.m_create(System)
                # TODO xyz does not contain cell info right?
                cell = cell if cell is None else atoms.get_cell() * ureg.angstrom
                sec_system.simulation_cell = cell
                sec_system.lattice_vectors = cell
                sec_system.configuration_periodic_dimensions = pbc
                sec_system.atom_positions = atoms.get_positions() * ureg.angstrom
                sec_system.atom_labels = atoms.get_chemical_symbols()
                sec_scc.single_configuration_calculation_to_system_ref = sec_system

        if len(sec_run.section_single_configuration_calculation) > 0:
            sec_scc = sec_run.section_single_configuration_calculation[-1]
            # final results are printed only in static/info
            # energies
            energies = self.info_parser.get('energies', {})
            for key, val in energies.items():
                metainfo_key = self._metainfo_keys_mapping.get(key, None)
                if metainfo_key is None:
                    continue
                setattr(sec_scc, metainfo_key, val)

            # forces
            forces = self.info_parser.get('forces')
            if forces is not None:
                sec_scc.atom_forces_free_raw = forces

            # eigenvalues
            eigenvalues = self.eigenvalue_parser.get('eigenvalues', self.info_parser.get('eigenvalues'))
            if eigenvalues is not None and eigenvalues.get('eigenvalues') is not None:
                kpts, eigs, occs = list(zip(*[e for e in eigenvalues.eigenvalues if e is not None]))
                eigs = np.transpose(eigs, axes=(2, 0, 1))
                occs = np.transpose(occs, axes=(2, 0, 1))
                sec_eigenvalues = sec_scc.m_create(Eigenvalues)
                if sec_run.section_method[-1].electronic_structure_method == 'DFT':
                    sec_eigenvalues.eigenvalues_kind = 'normal'
                if len(kpts) > 0:
                    sec_eigenvalues.eigenvalues_kpoints = kpts
                sec_eigenvalues.eigenvalues_values = eigs * self._units_mapping.get(self.info.get('energyunit').lower())
                sec_eigenvalues.eigenvalues_occupation = occs
                fermi_level = eigenvalues.get('fermi_energy')
                if fermi_level is not None:
                    unit = fermi_level.units
                    fermi_level = [fermi_level.magnitude] * len(eigs)
                    sec_scc.energy_reference_fermi = fermi_level * unit

            converged = self.out_parser.get('x_octopus_info_scf_converged_iterations')
            if converged is not None:
                sec_run.x_octopus_info_scf_converged_iterations = converged

            sec_scc.single_configuration_to_calculation_method_ref = sec_run.section_method[-1]
            sec_scc.single_configuration_calculation_to_system_ref = sec_run.section_system[-1]

    def parse_system(self):
        sec_run = self.archive.section_run[-1]
        sec_system = sec_run.m_create(System)

        cell = self.out_parser.get('grid', {}).get('cell')
        if cell is not None:
            sec_system.simulation_cell = cell
            sec_system.lattice_vectors = cell
        npbc = self.out_parser.get('grid', {}).get('npbc', 3)
        pbc = [True for _ in range(npbc)]
        sec_system.configuration_periodic_dimensions = pbc + [False] * (3 - len(pbc))

        symbols, coordinates = self.log_parser.get_coordinates()
        if len(coordinates) == 0:
            # get if from inp
            symbols, coordinates = self.inp_parser.get_coordinates()
        if len(coordinates) == 0:
            # try to read from file
            atoms = None
            file_types = {
                'PDBCoordinates': 'proteindatabank', 'XYZCoordinates': 'xyz',
                'XSFCoordinates': 'xsf'}
            for ftype, fformat in file_types.items():
                filename = self.info.get(ftype)
                if filename is None:
                    continue
                try:
                    atoms = read(os.path.join(self.maindir, filename), format=fformat)
                except Exception:
                    self.logger.error('Error reading coordinates file', data=dict(filename=filename))
                if atoms is not None:
                    symbols = atoms.get_chemical_symbols()
                    coordinates = atoms.get_positions()
                    break
        if len(coordinates) != 0:
            if self.info.get('ReducedCoordinates', None) is not None and cell is not None:
                coordinates = np.dot(coordinates, cell.magnitude)
                units = cell.units
            elif self.info.get('Coordinates', None) is not None:
                units = self.info.get('lengthunit')
            else:
                # read from ase atoms (in angstroms)
                units = 'angstrom'
            sec_system.atom_positions = coordinates * self._units_mapping.get(units.lower())
            sec_system.atom_labels = symbols
        else:
            self.logger.error('Error parsing atom positions and labels.')

    def parse_method(self):
        sec_run = self.archive.section_run[-1]
        sec_method = sec_run.m_create(Method)

        # smearing
        smearing_function = self._smearing_functions.get(self.info.get('SmearingFunction'), None)
        if smearing_function is not None:
            sec_method.smearing_kind = smearing_function
            smearing_width = self.info.get('Smearing', None)
            if smearing_width is not None:
                sec_method.smearing_width = (smearing_width * self._units_mapping.get(
                    self.info['energyunit'].lower())).to('joule').magnitude

        # check dft+u
        dft_u = self.info.get('DFTULevel')
        if dft_u is not None:
            sec_method.electronic_structure_method = 'DFT+U'
        else:
            theory_level = self._theory_levels.get(self.info.get('TheoryLevel', 4), None)
            if theory_level is not None:
                sec_method.electronic_structure_method = theory_level

        # xc functional
        xc_functionals = []
        xc_int_sum = 0
        xc_functional = self.info.get('XCFunctional', 0)
        if isinstance(xc_functional, str):
            xc_functionals = [xc.strip() for xc in xc_functional.upper().split('+')]
        else:
            # we try to resolve from output string
            for key in ['exchange', 'correlation']:
                val = self.out_parser.get('theory_level', {}).get(key, None)
                name, number = self._xc_functionals.get(val, (None, 0))
                if name is not None:
                    xc_functionals.append(name.upper())
                    xc_int_sum += number
            # get it from log
            diff = xc_functional - xc_int_sum
            if diff > 0:
                names_numbers = list(zip(*self._xc_functionals.values()))
                while diff > 0:
                    index = np.argmin(abs(np.array(names_numbers[1]) - diff))
                    xc_functionals.append(names_numbers[0][index].upper())
                    diff -= names_numbers[1][index]
                if diff < 0:
                    xc_functionals = []
                    self.logger.error(
                        'Error resolving xc functional', data=dict(XCfunctional=xc_functional))
        for xc_functional in xc_functionals:
            sec_xc = sec_method.m_create(XCFunctionals)
            sec_xc.XC_functional_name = xc_functional

        method_keys = ['SpinComponents', 'ExcessCharge']
        for key in method_keys:
            val = self.info.get(key)
            metainfo_key = self._metainfo_keys_mapping.get(key, None)
            if val is None or metainfo_key is None:
                continue
            try:
                setattr(sec_method, metainfo_key, int(val))
            except Exception:
                self.logger.warn('Error setting metainfo', data=dict(key=metainfo_key))

        # convert parsed values to proper metainfo type
        def convert_to_metainfo(prefix, key, val):
            metainfo_name = 'x_octopus_%s_%s' % (prefix, key)
            metainfo_type = self._metainfo_env.all_definitions_by_name.get(metainfo_name, None)
            if metainfo_type is not None:
                metainfo_type = metainfo_type[0].type
                if isinstance(metainfo_type, np.dtype):
                    val = np.array(val, metainfo_type)
                else:
                    val = metainfo_type(val)
            return metainfo_name, val

        # parse keywords from inp and parser.log
        sources = {'input': self.inp_parser, 'parserlog': self.log_parser}
        for prefix, source in sources.items():
            for key, val in source.info.items():
                try:
                    val = source.info.get(key)
                    metainfo_name, val = convert_to_metainfo(prefix, key, val)
                    setattr(sec_run, metainfo_name, val)
                except Exception:
                    self.logger.warn('Error setting metainfo', data=dict(key=key))

    def parse(self, filepath, archive, logger):
        self.filepath = filepath
        self.archive = archive
        self.logger = logging.getLogger(__name__) if logger is None else logger

        self.maindir = os.path.dirname(os.path.abspath(filepath))
        self.init_parser(filepath, logger)

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'Octopus'
        sec_run.program_basis_set_type = 'Real-space grid'

        sec_basis_set = sec_run.m_create(BasisSetCellDependent)
        sec_basis_set.basis_set_cell_dependent_kind = 'realspace_grids'

        header = self.out_parser.header
        sec_run.program_version = str(header.get('Version', ''))
        sec_run.x_octopus_log_svn_revision = int(header.get('Revision', 0))

        self.parse_method()

        self.parse_system()

        self.parse_scc()

        sec_workflow = archive.m_create(Workflow)
        sec_workflow.workflow_type = 'geometry_optimization'



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
The official NOMAD API.

There is a separate documentation for the API endpoints from a client perspective.

.. automodule:: nomad.app.flask.api.api
.. automodule:: nomad.app.flask.api.auth
.. automodule:: nomad.app.flask.api.upload
.. automodule:: nomad.app.flask.api.repo
.. automodule:: nomad.app.flask.api.archive
'''

from .api import api, blueprint
from . import info, auth, upload, repo, archive, encyclopedia, raw, mirror, dataset, metainfo



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Callable, Any, Dict, cast
import uuid
import numpy as np
import pint.quantity


from .metainfo import (
    Section, Quantity, MSection, MEnum, Datetime, Reference, Annotation, SectionAnnotation,
    DefinitionAnnotation)

'''
This module provides metainfo annotation class :class:`Elastic` and
class:`ElasticDocument` that allows to configure metainfo definitions for the use of
metainfo data in elastic search.
'''


class ElasticDocument(SectionAnnotation):
    '''
    This annotation class can be used to extend metainfo sections. It allows to detail
    how section instances (and their sub sections and quantities) should be represented in
    an elastic search index.

    In general sections with this annotation are mapped to elasticsearch_dsl document
    classes, sub sections become inner documents, and quantities with the :class:`Elastic`
    extension become fields in their respective document.

    Attributes:
        document: The elasticsearch_dsl document class that was generated from the
            metainfo section
    '''

    _all_documents: Dict[str, Any] = {}

    def __init__(self, index_name: str = None, id: Callable[[Any], str] = None):
        """
        Args:
            index_name: This is used to optionally add the index_name to the resulting
                elasticsearch_dsl document.
            id: A callable that produces an id from a section instance that is used as id
                for the respective elastic search index entry. The default will be randomly
                generated UUID(4).
        """
        self.index_name = index_name
        self.id = id

        self.fields: Dict[Quantity, str] = {}

    def new(self, section):
        return dict(elastic=ElasticEntry(section))

    @classmethod
    def create_index_entry(cls, section: MSection):
        ''' Creates an elasticsearch_dsl document instance for the given section. '''
        from elasticsearch_dsl import Object

        m_def = section.m_def
        annotation = m_def.m_get_annotations(ElasticDocument)
        document_cls = ElasticDocument._all_documents[m_def.qualified_name()]

        if annotation is None:
            obj = document_cls()
        else:
            if annotation.id is not None:
                id = annotation.id(section)
            else:
                id = uuid.uuid4()
            obj = document_cls(meta=dict(id=id))

        for quantity in m_def.all_quantities.values():
            for annotation in quantity.m_get_annotations(Elastic, as_list=True):
                if annotation.mapping is None:
                    continue

                value = annotation.value(section)
                if value is None or value == []:
                    continue

                # By default the full section is resolved for references
                if isinstance(quantity.type, Reference) and isinstance(annotation.mapping, Object):
                    if quantity.is_scalar:
                        value = ElasticDocument.create_index_entry(cast(MSection, value))
                    else:
                        value = [ElasticDocument.create_index_entry(item) for item in value]

                # Only the magnitude of scalar Pint quantity objects is stored
                if quantity.is_scalar and isinstance(value, pint.quantity._Quantity):
                    value = value.magnitude

                setattr(obj, annotation.field, value)

        for sub_section in m_def.all_sub_sections.values():
            try:
                if sub_section.repeats:
                    mi_values = list(section.m_get_sub_sections(sub_section))
                    if len(mi_values) == 0:
                        continue
                    value = [ElasticDocument.create_index_entry(value) for value in mi_values]
                else:
                    mi_value = section.m_get_sub_section(sub_section, -1)
                    if mi_value is None:
                        continue
                    value = ElasticDocument.create_index_entry(mi_value)

                setattr(obj, sub_section.name, value)
            except KeyError:
                # the sub section definition has no elastic quantities and therefore not
                # corresponding document class
                pass

        return obj

    @classmethod
    def index(cls, section: MSection, **kwargs):
        ''' Adds the given section to its elastic search index. '''
        entry = cls.create_index_entry(section)
        entry.save(**kwargs)
        return entry

    @property
    def document(self):
        return ElasticDocument.create_document(self.definition, index_name=self.index_name)

    @classmethod
    def create_document(
            cls, section: Section, attrs: Dict[str, Any] = None,
            prefix: str = None, index_name: str = None, root=True):
        '''
        Create all elasticsearch_dsl mapping classes for the section and its sub sections.
        '''
        document = cls._all_documents.get(section.qualified_name())
        if document is not None:
            return document

        from elasticsearch_dsl import Document, InnerDoc, Keyword, Date, Integer, Boolean, Object, Double, Float, Long, Nested

        if attrs is None:
            attrs = {}

        # create an field for each sub section
        for sub_section in section.all_sub_sections.values():
            sub_section_prefix = '%s.%s' % (prefix, sub_section.name) if prefix else sub_section.name

            inner_document = ElasticDocument.create_document(
                sub_section.sub_section, prefix=sub_section_prefix, index_name=index_name, root=False)
            if inner_document is not None:
                for annotation in sub_section.m_get_annotations(Elastic, as_list=True):
                    if annotation.nested:
                        assert sub_section.repeats, (
                            "Nested fields should be repeatable. If the subsection cannot be repeated, "
                            "define it as unnested instead."
                        )
                        attrs[sub_section.name] = Nested(inner_document)

                    annotation.register(prefix, annotation.field, index_name)

                if sub_section.name not in attrs:
                    attrs[sub_section.name] = Object(inner_document)

        # create an field for each quantity
        for quantity in section.all_quantities.values():
            first = True
            for annotation in quantity.m_get_annotations(Elastic, as_list=True):
                if annotation.mapping is None and first:
                    kwargs = dict(index=annotation.index)

                    # Find a mapping based on quantity type if not explicitly given
                    if quantity.type == str:
                        annotation.mapping = Keyword(**kwargs)
                    elif quantity.type in [float, np.float64] and quantity.is_scalar:
                        annotation.mapping = Double(**kwargs)
                    elif quantity.type == np.float32 and quantity.is_scalar:
                        annotation.mapping = Float(**kwargs)
                    elif quantity.type in [int, np.int32] and quantity.is_scalar:
                        annotation.mapping = Integer(**kwargs)
                    elif quantity.type == np.int64 and quantity.is_scalar:
                        annotation.mapping = Long(**kwargs)
                    elif quantity.type == bool:
                        annotation.mapping = Boolean(**kwargs)
                    elif quantity.type == Datetime:
                        annotation.mapping = Date(**kwargs)
                    elif isinstance(quantity.type, Reference):
                        inner_prefix = annotation.field
                        if prefix is not None:
                            inner_prefix = '%s.%s' % (prefix, inner_prefix)
                        inner_document = ElasticDocument.create_document(
                            cast(Section, quantity.type.target_section_def), prefix=inner_prefix, index_name=index_name, root=False)
                        annotation.mapping = Object(inner_document)
                    elif isinstance(quantity.type, MEnum):
                        annotation.mapping = Keyword(**kwargs)
                    else:
                        raise NotImplementedError(
                            'Quantity type %s for quantity %s is not supported.' % (quantity.type, quantity))

                assert first or annotation.mapping is None, 'Only the first Elastic annotation is mapped'

                if first:
                    assert annotation.field not in attrs, 'Elastic fields must be unique'
                    attrs[annotation.field] = annotation.mapping
                annotation.register(prefix, annotation.field, index_name)

                first = False

        if len(attrs) == 0:
            # do not create a document/inner document class, if no elastic quantities are defined
            return None

        doc_cls_obj = InnerDoc
        if root:
            doc_cls_obj = Document
        document = type(section.name, (doc_cls_obj,), attrs)
        cls._all_documents[section.qualified_name()] = document
        return document


class ElasticEntry(Annotation):
    def __init__(self, section: MSection):
        self.section = section

    def index(self, **kwargs):
        return ElasticDocument.index(self.section, **kwargs)

    def create_index_entry(self):
        return ElasticDocument.create_index_entry(self.section)


class Elastic(DefinitionAnnotation):
    '''
    This annotation class can be used to extend metainfo quantities. It allows to detail
    how this quantity should be represented in an elastic search index.

    Arguments:
        field:
            The name of the field for this quantity in the elastic search index. The
            default is the quantity name.
        mapping: A valid elasticsearch_dsl mapping. Default is ``Keyword()``.
        value:
            A callable that is applied to the containering section to get a value for
            this quantity when saving the section in the elastic search index. By default
            this will be the serialized quantity value.
        index:
            A boolean that indicates if this quantity should be indexed or merely be
            part of the elastic document ``_source`` without being indexed for search.
        aggregateable:
            A boolean that determines, if this quantity can be used in aggregations
    '''
    def __init__(
            self,
            field: str = None,
            mapping: Any = None,
            value: Callable[[Any], Any] = None,
            index: bool = True):

        self.field = field
        self.mapping = mapping
        self.value = value
        self.index = index

        self.prefix = None
        self.qualified_field = field

    def init_annotation(self, definition):
        super().init_annotation(definition)

        if self.field is None:
            self.field = definition.name

        if self.value is None:
            self.value = lambda section: section.m_get(definition)

    def register(self, prefix: str, field: str, index: str):

        if prefix is None:
            self.qualified_field = field
        else:
            self.qualified_field = '%s.%s' % (prefix, field)

    @property
    def aggregateable(self):
        return self.mapping is None or self.mapping.__class__.__name__ == 'Keyword'



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class Dmol3Parser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*\-*\+*\d*'

        super().__init__(
            specifications=dict(
                name='parsers/dmol', code_name='DMol3',
                code_homepage='http://dmol3.web.psi.ch/dmol3.html', domain='dft',
                mainfile_name_re=r'.*\.outmol',
                mainfile_contents_re=r'Materials Studio DMol\^3'),
            units_mapping=dict(length=ureg.bohr, energy=ureg.hartree),
            program_version=r'Materials Studio DMol\^3 (version \S+)',
            atom_labels_atom_positions=r'ATOMIC  COORDINATES \(au\) +DERIVATIVES \(au\)([\s\S]+?)\n *\n',
            energy_total=rf'Total E \(au\).+\s*\-+\s*Ef +({re_f}).+\s*\-+')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
This module contains all functions necessary to manage DOI via datacite.org and its
MDS API (https://support.datacite.org/docs/mds-api-guide).
'''
import xml.etree.ElementTree as ET
import datetime
import requests
from requests.auth import HTTPBasicAuth
from mongoengine import Document, StringField, DateTimeField
from mongoengine.errors import NotUniqueError

from nomad.datamodel import User
from nomad import config, utils


def edit_url(doi: str, url: str = None):
    ''' Changes the URL of an already findable DOI. '''
    if url is None:
        url = 'https://nomad-lab.eu/prod/rae/gui/datasets/doi/%s' % doi

    metadata_url = '%s/doi/%s' % (config.datacite.mds_host, doi)
    response = requests.put(
        metadata_url,
        headers={'Content-Type': 'text/plain;charset=UTF-8'},
        data='doi=%s\nurl=%s' % (doi, url), **_requests_args())

    if response.status_code >= 300:
        raise Exception('Unexpected datacite response (status code %d): %s' % (
            response.status_code, response.text))


def _xml(parent, element: str, value: str = None):
    path = element.split('/')
    el = parent
    for segment in path:
        el = ET.SubElement(el, segment)

    if value is not None:
        el.text = value

    return el


def _requests_args():
    return dict(auth=HTTPBasicAuth(config.datacite.user, config.datacite.password))


class DOI(Document):
    doi = StringField(primary_key=True)
    url = StringField()
    metadata_url = StringField()
    doi_url = StringField()
    state = StringField()
    create_time = DateTimeField()
    metadata_xml = StringField()

    @staticmethod
    def create(title: str, user: User) -> 'DOI':
        ''' Creates a unique DOI with the NOMAD DOI prefix. '''
        # TODO We use a collection of all DOIs in mongo to ensure uniqueness. We attempt
        # to create new DOIs based on a counter per day until we find a non existing DOI.
        # This might be bad if many DOIs per day are to be expected.
        counter = 1
        create_time = datetime.datetime.now()

        while True:
            doi_str = '%s/NOMAD/%s-%d' % (
                config.datacite.prefix, create_time.strftime('%Y.%m.%d'), counter)

            try:
                doi = DOI(doi=doi_str)
                doi.save(force_insert=True)
                break
            except NotUniqueError:
                counter += 1

        doi.metadata_url = '%s/metadata/%s' % (config.datacite.mds_host, doi_str)
        doi.doi_url = '%s/doi/%s' % (config.datacite.mds_host, doi_str)
        doi.state = 'created'
        doi.create_time = create_time
        doi.url = '%s/dataset/doi/%s' % (config.gui_url(), doi_str)

        affiliation = ''
        if user.affiliation is not None:
            affiliation += user.affiliation.strip()
        if user.affiliation_address is not None:
            affiliation += '; ' + user.affiliation_address.strip()

        if title is None or title.strip() == '':
            title = 'NOMAD Repository Dataset'

        mds_resource = ET.Element("resource")
        mds_resource.attrib['xsi:schemaLocation'] = 'http://datacite.org/schema/kernel-3 http://schema.datacite.org/meta/kernel-3.1/metadata.xsd'
        mds_resource.attrib['xmlns'] = 'http://datacite.org/schema/kernel-3'
        mds_resource.attrib['xmlns:xsi'] = 'http://www.w3.org/2001/XMLSchema-instance'

        mds_identifier = ET.SubElement(mds_resource, 'identifier')
        mds_identifier.text = doi_str
        mds_identifier.attrib['identifierType'] = 'DOI'

        mds_creator = _xml(mds_resource, 'creators/creator')
        _xml(mds_creator, 'creatorName', user.name)
        if affiliation != '':
            _xml(mds_creator, 'affiliation', affiliation)
        _xml(mds_resource, 'titles/title', title.strip())
        _xml(mds_resource, 'publisher', 'NOMAD Repository')
        _xml(mds_resource, 'publicationYear', str(datetime.datetime.now().year))

        doi.metadata_xml = ET.tostring(mds_resource, encoding='UTF-8', method='xml').decode('utf-8')
        doi.save()

        return doi

    def __handle_datacite_errors(self, response, msg: str):
        if response.status_code >= 300:
            utils.get_logger(__name__).error(
                'could not %s' % msg,
                status_code=response.status_code, body=response.content,
                doi=self.doi)

            return False
        else:
            return True

    def create_draft(self):
        if config.datacite.enabled:
            assert self.state == 'created', 'can only create a draft for created DOIs'
            response = requests.post(
                self.metadata_url,
                headers={'Content-Type': 'application/xml;charset=UTF-8'},
                data=self.metadata_xml.encode('utf-8'), **_requests_args())

            if self.__handle_datacite_errors(response, 'create draft DOI'):
                self.state = 'draft'
                self.save()

    def delete(self, *args, **kwargs):
        if config.datacite.enabled:
            assert self.state == 'draft', 'can only delete drafts'
            response = requests.delete(self.metadata_url, **_requests_args())

            self.__handle_datacite_errors(response, 'delete draft DOI')

        super().delete(*args, **kwargs)

    def make_findable(self):
        if config.datacite.enabled:
            assert self.state == 'draft', 'can only make drafts findable'
            body = ('doi=%s\nurl=%s' % (self.doi, self.url)).encode('utf-8')
            response = requests.put(
                self.doi_url, **_requests_args(),
                headers={'Content-Type': 'text/plain;charset=UTF-8'}, data=body)

            if self.__handle_datacite_errors(response, 'make DOI findable'):
                self.state = 'findable'
                self.save()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
import re
from ase.cell import Cell

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class TinkerParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        def get_positions(val):
            res = dict(lattice_vectors=[], atom_labels=[], atom_positions=[])
            try:
                cell_par = re.search(rf'({re_f} +{re_f} +{re_f} +{re_f} +{re_f} +{re_f})', val)
                res['lattice_vectors'] = Cell.fromcellpar([float(v) for v in cell_par.group(1).split()]).array * ureg.angstrom
                labels, positions = zip(*re.findall(rf'\d+ +([A-Z][a-z]*) +({re_f} +{re_f} +{re_f} +)\d+', val))
                res['atom_labels'] = labels
                res['atom_positions'] = np.array([v.split() for v in positions], dtype=np.dtype(np.float64)) * ureg.angstrom
            except Exception:
                pass
            return res

        super().__init__(
            specifications=dict(
                name='parsers/tinker', code_name='TINKER', domain='dft',
                mainfile_contents_re=r'TINKER  ---  Software Tools for Molecular Design'),
            units_mapping=dict(energy=ureg.kcal / 6.02214076e+23, length=ureg.angstrom),
            program_version=r'\# +(Version [\d\.]+)',
            auxilliary_files=r'Coordinate File +(\S+)',
            atom_labels_atom_positions_lattice_vectors=(
                rf'\d+ +\w+ +.+\s*({re_f} +{re_f} +{re_f} +{re_f} +{re_f} +{re_f}[\s\S]+?)(?:[\(,]|\Z)',
                get_positions),
            # TODO support for various input, output formats
            energy_total=rf'Total Energy +({re_f}) Kcal\/mole')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import numpy as np

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class NAMDParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*e*\-*\+*\d*'

        def get_positions(val):
            try:
                labels, positions = zip(*re.findall(rf'ATOM +\d+ +([A-Z])\w* +\w+ +\d+ +({re_f} +{re_f} +{re_f})', val))
                positions = np.array([v.split() for v in positions], dtype=np.dtype(np.float64))
                return dict(atom_labels=labels, atom_positions=positions)
            except Exception:
                return dict()

        super().__init__(
            specifications=dict(
                name='parsers/namd', code_name='Namd', domain='dft',
                mainfile_contents_re=r'\s*Info:\s*NAMD\s*[0-9.]+\s*for\s*',
                mainfile_mime_re=r'text/.*'),
            units_mapping=dict(length=ureg.angstrom, energy=ureg.kcal / 6.02214076e+23),
            auxilliary_files=r'Info\: COORDINATE PDB +(\S+\.pdb)',
            # due to auto type conversion, we need to include NAMD to make it string
            program_version=r'Info\: (NAMD [\d\.]+) for',
            atom_labels_atom_positions=(rf'(ATOM +\d+ +\w+ +\w+ +\d+ +{re_f}[\s\S]+?)END', get_positions),
            energy_total=(rf'ENERGY\: +(\d+ +{re_f}.+)', lambda x: x.split()[10]))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from typing import Tuple, Generator, cast

from nomad.metainfo.metainfo import Quantity, Reference
from nomad.metainfo.search_extension import Search
from nomad.search import search_quantities


def provider_specific_fields() -> Generator[Tuple[str, Search], None, None]:
    for search_quantity in search_quantities.values():
        quantity = cast(Quantity, search_quantity.definition)
        if isinstance(quantity.type, Reference):
            # we can't yet support those
            continue

        nmd_name = search_quantity.qualified_name
        nmd_name_split = nmd_name.split('.')

        if len(nmd_name_split) == 1:
            # plain metadata
            pass
        elif not nmd_name_split[0] in ['dft', 'encyclopedia']:
            # other domains fields that do not make sense in the optimade context
            continue
        elif len(nmd_name_split) > 2 and nmd_name_split[1] == 'optimade':
            # these are already in optimade
            continue

        opt_name = nmd_name.replace('.', '_')
        yield opt_name, search_quantity



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from abc import ABC, abstractmethod
from collections import OrderedDict
import numpy as np
from typing import Tuple, List
from nomad.units import ureg

from nomad.metainfo import Section
from nomad.utils import RestrictedDict


def get_basis_set(context, entry_archive, logger) -> RestrictedDict:
    """Decide which type of basis set settings are applicable to the entry and
    return a corresponding settings as a RestrictedDict.

    Args:
        context: The calculation context.
        entry_archive: EntryArchive from which values are extracted.
        logger: Shared logger.

    Returns:
        RestrictedDict or None: Returns the extracted settings as a
        RestrictedDict. If no suitable basis set settings could be identified,
        returns None.
    """
    settings: BasisSet = None
    program_name = entry_archive.section_run[0].program_name
    if program_name == "exciting":
        settings = BasisSetExciting(context, entry_archive, logger)
    elif program_name == "FHI-aims":
        settings = BasisSetFHIAims(context, entry_archive, logger)
    else:
        return None

    return settings.to_dict()


class BasisSet(ABC):
    """Abstract base class for basis set settings. The idea is to create
    subclasses that inherit this class and hierarchically add new mandatory and
    optional settings with the setup()-function.
    """
    def __init__(self, context, entry_archive, logger):
        """
        """
        self._ctx = context
        self._entry_archive = entry_archive
        self._logger = logger
        mandatory, optional = self.setup()
        self.settings = RestrictedDict(mandatory, optional, forbidden_values=[None])

    @abstractmethod
    def to_dict(self) -> RestrictedDict:
        """Used to extract basis set settings from the archive and returning
        them as a RestrictedDict.
        """
        pass

    @abstractmethod
    def setup(self) -> Tuple:
        """Used to define a list of mandatory and optional settings for a
        subclass.

        Returns:
            Should return a tuple of two lists: the first one defining
            mandatory keys and the second one defining optional keys.
        """
        mandatory: List = []
        optional: List = []
        return mandatory, optional


class BasisSetFHIAims(BasisSet):
    """Basis set settings for 'FHI-Aims' (code-dependent).
    """
    def setup(self) -> Tuple:
        # Get previously defined values from superclass
        mandatory, optional = super().setup()

        # Add new values
        mandatory += ["fhiaims_basis"]

        return mandatory, optional

    def to_dict(self):
        # Get basis set settings for each species
        aims_bs = self._ctx.representative_method.x_fhi_aims_section_controlIn_basis_set
        if aims_bs is not None:
            bs_by_species = {}
            for this_aims_bs in aims_bs:
                this_bs_dict = self._values_to_dict(this_aims_bs, level=2)
                this_species = this_aims_bs['x_fhi_aims_controlIn_species_name'][0]
                bs_by_species[this_species] = this_bs_dict

            # Sort alphabetically by species label
            if bs_by_species:
                basis = OrderedDict()
                for k in sorted(bs_by_species.keys()):
                    basis[k] = bs_by_species[k]
                self.settings["fhiaims_basis"] = basis

        return self.settings

    @classmethod
    def _values_to_dict(cls, data, level=0):
        result = None
        if data is None:
            return None
        elif isinstance(data, (Section, dict)):
            result = OrderedDict()
            for k in sorted(cls._filtered_section_keys(data)):
                v = data.get(k, None)
                result[k] = cls._values_to_dict(v, level=level + 1)
        elif isinstance(data, (list)):
            result = []
            for k in range(len(data)):
                v = data[k]
                result.append(cls._values_to_dict(v, level=level + 1))
        elif isinstance(data, (np.ndarray)):
            result = data.tolist()
        else:
            result = data
        return result

    @classmethod
    def _filtered_section_keys(cls, section):
        for k in section.keys():
            # skip JSON-specific keys
            if k == '_gIndex':
                continue
            if k == '_name':
                continue
            else:
                # json values and subsections
                yield k


class BasisSetExciting(BasisSet):
    """Basis set settings for 'Exciting' (code-dependent).
    """
    def setup(self) -> Tuple:
        # Get previously defined values from superclass
        mandatory, optional = super().setup()

        # Add new values
        mandatory += [
            "muffin_tin_settings",
            "rgkmax",
            "gkmax",
            "lo",
            "lmaxapw",
        ]

        return mandatory, optional

    def to_dict(self):
        """Special case of basis set settings for Exciting code. See list at:
        https://gitlab.mpcdf.mpg.de/nomad-lab/encyclopedia-general/wikis/FHI-visit-preparation
        """
        # Add the muffin-tin settings for each species ordered alphabetically by atom label
        try:
            groups = self._ctx.representative_system.x_exciting_section_atoms_group
            groups = sorted(groups, key=lambda group: group.x_exciting_geometry_atom_labels)
            muffin_tin_settings = OrderedDict()
            for group in groups:
                label = group.x_exciting_geometry_atom_labels
                try:
                    muffin_tin_settings["{}_muffin_tin_radius".format(label)] = "%.6f" % (group.x_exciting_muffin_tin_radius.to(ureg.angstrom).magnitude)
                except Exception:
                    muffin_tin_settings["{}_muffin_tin_radius".format(label)] = None
                try:
                    muffin_tin_settings["{}_muffin_tin_points".format(label)] = "%d" % group.x_exciting_muffin_tin_points
                except Exception:
                    muffin_tin_settings["{}_muffin_tin_points".format(label)] = None
            self.settings["muffin_tin_settings"] = muffin_tin_settings
        except Exception:
            pass

        # Other important method settings
        system = self._ctx.representative_system
        try:
            self.settings['rgkmax'] = "%.6f" % (system.x_exciting_rgkmax.magnitude)
        except Exception:
            pass
        try:
            self.settings['gkmax'] = "%.6f" % (1e-10 * system.x_exciting_gkmax.magnitude)
        except Exception:
            pass
        try:
            self.settings['lo'] = "%d" % (system.x_exciting_lo)
        except Exception:
            pass
        try:
            self.settings['lmaxapw'] = "%d" % (system.x_exciting_lmaxapw)
        except Exception:
            pass

        return self.settings



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
import re
import ase
from ase import io as aseio

from .metainfo import m_env
from nomad.units import ureg
from nomad.parsing.parser import FairdiParser
from nomad.parsing.file_parser import TextParser, Quantity, FileParser, DataTextParser
from nomad.datamodel.metainfo.common_dft import Run, Method, System, XCFunctionals,\
    MethodAtomKind, BasisSetCellDependent, BasisSetAtomCentered, MethodBasisSet,\
    SingleConfigurationCalculation, ScfIteration, SamplingMethod

from .metainfo.cp2k_general import x_cp2k_section_quickstep_settings, x_cp2k_section_dbcsr,\
    x_cp2k_section_startinformation, x_cp2k_section_end_information, x_cp2k_section_program_information,\
    x_cp2k_section_global_settings, x_cp2k_section_vdw_settings, x_cp2k_section_atomic_kinds,\
    x_cp2k_section_atomic_kind, x_cp2k_section_kind_basis_set, x_cp2k_section_total_numbers,\
    x_cp2k_section_maximum_angular_momentum, x_cp2k_section_quickstep_calculation,\
    x_cp2k_section_scf_iteration, x_cp2k_section_stress_tensor, x_cp2k_section_md_settings,\
    x_cp2k_section_md_step, x_cp2k_section_restart_information, x_cp2k_section_geometry_optimization,\
    x_cp2k_section_geometry_optimization_step


try:
    import mdtraj
except ImportError:
    logging.getLogger(__name__).warn('Required MDTraj module not found.')
    mdtraj = False


units_map = {
    'hbar': ureg.hbar, 'hartree': ureg.hartree, 'angstrom': ureg.angstrom,
    'au_t': ureg.hbar / ureg.hartree}


def resolve_unit(unit_str, parts=[]):
    unit_str = unit_str.lower().replace(' ', '')
    parts = list(parts)

    if unit_str in units_map:
        return units_map[unit_str]

    try:
        return float(unit_str)
    except Exception:
        pass

    if unit_str == '':
        return 1

    open_p = unit_str.rfind('(')
    if open_p > -1:
        n_groups = unit_str.count('(')
        if n_groups != unit_str.count(')'):
            return unit_str
        for n in range(n_groups):
            part = unit_str[open_p + 1:]
            part = part[:part.find(')')]
            parts.append(resolve_unit(part, parts))
            unit_str = unit_str.replace('(%s)' % part, '[%d]' % n)
            open_p = unit_str.rfind('(')
        return resolve_unit(unit_str, parts)

    vals = unit_str.split('/')
    if len(vals) > 1:
        vals = [resolve_unit(v, parts) for v in vals]
        val = vals[0]
        for v in vals[1:]:
            val /= v
        return val

    vals = unit_str.split('**')
    if len(vals) > 1:
        vals = [resolve_unit(v, parts) for v in vals]
        val = vals[0]
        for v in reversed(vals[1:]):
            val = val ** v
        return val

    vals = unit_str.split('^')
    if len(vals) > 1:
        vals = [resolve_unit(v, parts) for v in vals]
        val = vals[0]
        for v in reversed(vals[1:]):
            val = val ** v
        return val

    vals = unit_str.split('*')
    if len(vals) > 1:
        vals = [resolve_unit(v, parts) for v in vals]
        unit = 1
        for v in vals:
            unit *= v
        return unit

    vals = unit_str.split('-1')
    if len(vals) == 2:
        return 1 / resolve_unit(vals[0], parts)

    vals = re.match(r'\[(\d+)\]', unit_str)
    if vals:
        return parts[int(vals.group(1))]


class Property:
    def __init__(self, **kwargs):
        self._data = kwargs

    def __getattr__(self, key):
        return self._data.get(key, None)


class Trajectory(Property):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class XYZTrajParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):

        def get_trajectory(val_in):
            val = np.transpose([v.split() for v in val_in.strip().split('\n')])
            positions = np.array(val[1:4], dtype=float).T
            return Trajectory(labels=val[0], positions=positions)

        self.quantities = [
            Quantity(
                'trajectory', r'([A-Z][a-z]?[\w\.\-\s]+?)(?:\s+\d\n|\Z)',
                str_operation=get_trajectory, repeats=True),
            Quantity(
                'energy', r'E\s*=\s*(\S+)', repeats=True, dtype=float)]


class TrajParser(FileParser):
    def __init__(self, **kwargs):
        super().__init__()
        self._xyz_parser = XYZTrajParser()
        self.format = None
        self.units = None
        self.type = kwargs.get('type', 'positions')

    @property
    def trajectory(self):
        if self._file_handler is None:
            if self.format is None:
                self.format = self.mainfile.split('.')[-1].lower()

            result = None
            labels = []
            # ase is better as it reads also symbols
            try:
                atoms_list = [atoms for atoms in aseio.iread(self.mainfile, format=self.format)]
                result = [atoms.positions for atoms in atoms_list]
                labels = [list(atoms.symbols) for atoms in atoms_list]

            # custom parser
            except Exception:
                if self.format == 'xyz':
                    self._xyz_parser.mainfile = self.mainfile
                    result = [traj.positions for traj in self._xyz_parser.get('trajectory')]
                    labels = [traj.labels for traj in self._xyz_parser.get('trajectory')]

            if result is None and mdtraj:
                reader = None
                if self.format in ['xyz', 'xmol', 'atomic']:
                    reader = mdtraj.formats.XYZTrajectoryFile(self.mainfile)
                elif self.format == 'dcd':
                    reader = mdtraj.formats.DCDTrajectoryFile(self.mainfile)
                elif self.format == 'pdb':
                    reader = mdtraj.formats.PDBTrajectoryFile(self.mainfile)
                else:
                    self.logger.error('Unsupported trajectory format.')

                if reader is not None:
                    try:
                        # we do not stream to simplify archive writing
                        result = reader.read()
                    except Exception:
                        pass

            result = result * self.units if self.units is not None else result

            result = [Trajectory(**{self.type: res}) for res in result]

            # add labels to trajectory
            for n, labels_i in enumerate(labels):
                result[n]._data.update({'labels': labels_i})

            self._file_handler = result

        return self._file_handler


class ForceParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_float = r'[\d\.\-\+eE]+'
        self._quantities = [Quantity(
            'atom_forces',
            rf'\d+\s*\d+\s*\w+\s*({re_float})\s*({re_float})\s*({re_float})', repeats=True)]


class XCFunctional(Property):
    def __init__(self, name, **kwargs):
        super().__init__(name=name, **kwargs)


class InpValue:
    def __init__(self, name, **kwargs):
        self._data = kwargs
        self._name = name
        self._dict = None

    @property
    def name(self):
        return self._name

    def add(self, key, val):
        if key in self._data:
            self._data[key] = self._data[key] if isinstance(self._data[key], list) else [self._data[key]]
            self._data[key].append(val)
        else:
            self._data[key] = val

    def to_dict(self):
        if self._dict is None:
            def extract(data):
                out = dict()
                for key, val in data.items():
                    if isinstance(val, InpValue):
                        val = extract(val.to_dict())
                    out[key] = val
                return out

            self._dict = extract(self._data)

        return self._dict

    def items(self):
        for key, val in self._data.items():
            yield key, val

    def __getattr__(self, key):
        return self._data.get(key, None)

    def __repr__(self):
        return self._name


class InpParser(FileParser):
    def __init__(self):
        super().__init__()
        self._re_open = re.compile(r'&(\w+)\s*(.*)[#!]*')
        self._re_close = re.compile(r'&END')
        self._re_key_value = re.compile(r'(\w+)\s+(.+)[#!]*')
        self._re_set_variable = re.compile(r'\@SET\s+(\w+)\s+(.+)')
        self._re_variable = re.compile(r'\$\{\w+\}')

    @property
    def tree(self):
        if self._file_handler is None:
            def override(data):
                if data[0] == 'PROJECT':
                    return 'PROJECT_NAME', data[1]
                elif not data[0].isupper():
                    return 'DEFAULT_KEYWORD', ' '.join(data)
                return data

            line = True
            sections = [InpValue('tree')]
            variables = dict()
            while line:
                line = self.mainfile_obj.readline()
                # comments
                strip = line.strip()
                if not strip or strip[0] in ('#', '!'):
                    continue
                set_variable = self._re_set_variable.search(line)
                if set_variable:
                    variables['${%s}' % set_variable.group(1)] = set_variable.group(2)
                    continue
                close_section = self._re_close.search(line)
                if close_section:
                    sections.pop(-1)
                    continue
                open_section = self._re_open.search(line)
                if open_section:
                    section = InpValue(open_section.group(1))
                    sections[-1].add(open_section.group(1), section)
                    sections.append(section)
                    if open_section.group(2):
                        sections[-1].add('VALUE', open_section.group(2))
                    continue
                key_value = self._re_key_value.search(line)
                if key_value:
                    key_value = list(key_value.groups())
                    key_value[1] = variables.get(key_value[1], key_value[1])
                    key_value = override(key_value)
                    sections[-1].add(key_value[0], key_value[1])
                    continue
            self._file_handler = sections[0]
        return self._file_handler

    def parse(self, key):
        if self._results is None:
            self._results = dict()

        source = self.tree.to_dict()
        for sub_key in key.strip('/').split('/'):
            val = source.get(sub_key, None)
            source = val
            if val is None:
                break
        self._results[key] = val


class CP2KOutParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_header(val_in):
            val = val_in.split('  ', 1)
            return [val[0].strip().replace(' ', '_').lower(), val[-1].strip()]

        def str_to_program(val_in):
            val = val_in.split(' ', 2)
            return ['_'.join(val[:2]).lower(), ''.join([v.strip() for v in val[2].split('\n')])]

        def str_to_atomic_coordinates(val_in):
            val = [v.split() for v in val_in.split('\n')]
            lengthunit = val[0][0].lower()
            val = np.transpose(np.array([v for v in val if len(v) == 9]))
            labels = val[2]
            positions = np.transpose(np.array(val[4:7], dtype=float)) * resolve_unit(lengthunit)
            atomic_numbers = {element: int(val[3][n]) for n, element in enumerate(val[2])}
            return Trajectory(labels=labels, positions=positions, atomic_numbers=atomic_numbers)

        def str_to_stress_eigenvalues(val_in):
            val = [v.split() for v in val_in.strip().split('\n')]
            val = np.array([v for v in val if v], dtype=float)
            return val[0] * ureg.GPa, val[1:]

        def str_to_iteration(val_in):
            val = val_in.strip().split()
            return {
                'energy_total_scf_iteration': float(val[-2]) * ureg.hartree,
                'energy_change_scf_iteration': float(val[-1]) * ureg.hartree}

        def str_to_information(val_in):
            val = [v.split('=') for v in val_in.strip().split('\n')]
            return {v[0].strip().lower().replace(' ', '_').replace('.', ''): v[1] for v in val if len(v) == 2}

        re_float = r'[\d\.\-\+eE]+'

        n_orbital_basis_quantities = [Quantity(
            'basis_set_number_of_%s' % key.lower().replace(' ', '_'),
            r'Number of %s:\s+(\d+)' % key, dtype=int) for key in [
                'orbital shell sets', 'orbital shells', 'primitive Cartesian functions',
                'Cartesian basis functions', 'spherical basis functions']]

        energy_quantities = [Quantity(
            '%s' % key.lower().replace(' ', '_').replace('-', '_'),
            rf'%s:\s*({re_float})' % key, dtype=float, unit='hartree', repeats=True) for key in [
                'Hartree energy', 'Exchange-correlation energy', 'Electronic kinetic energy',
                'Total energy']]
        # what is the difference between Total energy and ENERGY| Total

        scf_wavefunction_optimization_quantities = [
            Quantity(
                'iteration',
                rf'(\d+\s+\S+\s*\S*\s+{re_float}\s+{re_float}\s+{re_float}\s+{re_float}\s+{re_float})\n',
                repeats=True, convert=False, str_operation=str_to_iteration),
            # TODO add minimizer info
            Quantity(
                'converged',
                r'SCF run converged in\s*(\d+) steps', dtype=int),
            # find example with cueb file
            Quantity(
                'cube_file',
                r' The electron density is written in cube file format to the file:\s*(.+?\.cube)'),
            # TODO add mulliken, hirschfield
            Quantity(
                'energy_total',
                rf'ENERGY\| Total FORCE_EVAL \( \w+ \) energy \(a\.u\.\):\s*({re_float})',
                dtype=float, unit='hartree'),
            Quantity(
                'atom_forces',
                rf'ATOMIC FORCES in \[a\.u\.\]\s*.+([\s\S]+?)SUM', convert=False,
                str_operation=lambda x: np.array(np.transpose(
                    [v.split() for v in x.strip().split('\n')])[3:6], dtype=float).T),
            # TODO test stress cannot find example
            Quantity(
                'stress_tensor',
                r' (?:NUMERICAL )?STRESS TENSOR \[GPa\]\s+X\s+Y\s+Z\s+([\d\.\-\s]+)',
                str_operation=lambda x:np.array([v.split() for v in x.strip().split('\n')], dtype=float),
                convert=False, unit='GPa'),
            Quantity(
                'stress_tensor_one_third_of_trace',
                rf'  1/3 Trace\(stress tensor\)\s*:\s*({re_float})', dtype=float, unit='GPa'),
            Quantity(
                'stress_tensor_determinant',
                rf'Det\(stress tensor\)\s*:\s*({re_float})', dtype=float, unit='GPa**3'),
            Quantity(
                'stress_eigenvalues_eigenvectors',
                r' EIGENVECTORS AND EIGENVALUES OF THE STRESS TENSOR\s*([\d\.\-\s]+)',
                str_operation=str_to_stress_eigenvalues, convert=False)] + energy_quantities

        single_point_quantities = [
            Quantity(
                'self_consistent',
                r'SCF WAVEFUNCTION OPTIMIZATION([\s\S]+?)\Z', repeats=False,
                sub_parser=TextParser(quantities=scf_wavefunction_optimization_quantities)),
            # TODO add rpa, etc.
        ]

        geometry_optimization_quantities = [
            Quantity(
                'method',
                r'\*{3}\s*((?:CONJUGATE GRADIENTS|L-BFGS|BFGS))\s*\*{3}', flatten=False),
            Quantity(
                'self_consistent',
                r'SCF WAVEFUNCTION OPTIMIZATION([\s\S]+?)OPTIMIZ', repeats=False,
                sub_parser=TextParser(quantities=scf_wavefunction_optimization_quantities)),
            Quantity(
                'optimization_step',
                r'(ATION STEP:\s*\d+[\s\S]+?)(?:\-\s+OPTIMIZ|\Z)', repeats=True,
                sub_parser=TextParser(quantities=[
                    # TODO parse atomic positions
                    Quantity('step', r'ATION STEP:\s*(\d+)'),
                    # I do not quite get why there can be multiple scfs in a step
                    Quantity(
                        'information',
                        r'Informations at step\s*=\s*\d+\s*\-+([\s\S]+?)\-{5}',
                        str_operation=str_to_information),
                    Quantity(
                        'self_consistent',
                        r'FUNCTION OPTIMIZATION([\s\S]+?)(?: SCF WAVE|\Z)', repeats=True,
                        sub_parser=TextParser(quantities=scf_wavefunction_optimization_quantities))]))
        ]

        molecular_dynamics_quantities = [
            Quantity(
                'initial',
                r' INITIAL\| (.+? {2})=\s+(.+)', str_operation=str_to_header, repeats=True),
            Quantity(
                'self_consistent',
                r'SCF WAVEFUNCTION OPTIMIZATION([\s\S]+?)(?:\*\n *ENSEM|\Z)', repeats=False,
                sub_parser=TextParser(quantities=scf_wavefunction_optimization_quantities)),
            Quantity(
                'md_step',
                r'(BLE TYPE[\s\S]+?)(?:\*\n *ENSEM|\Z)',
                repeats=True, sub_parser=TextParser(quantities=[
                    Quantity(
                        'ensemble_type', r'BLE TYPE\s*=\s*(.+)'),
                    Quantity(
                        'step_number', r'STEP NUMBER\s*=\s*(\d+)', dtype=int),
                    Quantity(
                        'time', rf'TIME \[fs\]\s*=\s*({re_float})', dtype=float),
                    Quantity(
                        'conserved_quantity',
                        rf'CONSERVED QUANTITY \[hartree\]\s*=\s*({re_float})',
                        dtype=float, unit='hartree'),
                    Quantity(
                        'cpu_time',
                        rf'CPU TIME \[s\]\s*=\s*({re_float})\s*(re_float)',
                        dtype=float),
                    Quantity(
                        'energy_drift',
                        rf'ENERGY DRIFT PER ATOM \[K\]\s*=\s*({re_float})\s*({re_float})',
                        dtype=float),
                    Quantity(
                        'potential_energy',
                        rf'POTENTIAL ENERGY\[hartree\]\s*=\s*({re_float})\s*({re_float})',
                        dtype=float, unit='hartree'),
                    Quantity(
                        'kinetic_energy',
                        rf'KINETIC ENERGY\[hartree\]\s*=\s*({re_float})\s*({re_float})',
                        dtype=float, unit='hartree'),
                    Quantity(
                        'temperature',
                        rf'TEMPERATURE \[K\]\s*=\s*({re_float})\s*({re_float})', dtype=float),
                    Quantity(
                        'pressure',
                        rf'PRESSURE \[bar\]\s*=\s*({re_float})\s*({re_float})',
                        dtype=float, unit='bar'),
                    Quantity(
                        'barostat_temperature',
                        rf'BAROSTAT TEMP\[K\]\s*=\s*({re_float})\s*({re_float})', dtype=float),
                    Quantity(
                        'volume',
                        rf'VOLUME\[bohr\^3\]\s*=\s*({re_float})\s*({re_float})',
                        dtype=float, unit='bohr**3'),
                    Quantity(
                        'cell_length_instantaneous',
                        rf'CELL LNTHS\[bohr\]\s*=\s*({re_float})\s*({re_float})\s*({re_float})',
                        dtype=float, unit='bohr'),
                    Quantity(
                        'cell_length_average',
                        rf'AVE\. CELL LNTHS\[bohr\]\s*=\s*({re_float})\s*({re_float})\s*({re_float})',
                        dtype=float, unit='bohr'),
                    Quantity(
                        'cell_angle_instantaneous',
                        rf'CELL ANGLS\[deg\]\s*=\s*({re_float})\s*({re_float})\s*({re_float})',
                        dtype=float),
                    Quantity(
                        'cell_angle_average',
                        rf'AVE\. CELL ANGLS\[deg\]\s*=\s*({re_float})\s*({re_float})\s*({re_float})',
                        dtype=float),
                    Quantity(
                        'self_consistent',
                        r'FUNCTION OPTIMIZATION([\s\S]+?)(?: SCF WAVE|\Z)', repeats=True,
                        sub_parser=TextParser(quantities=scf_wavefunction_optimization_quantities))])
            )
        ]

        quickstep_quantities = [
            Quantity(
                'dft',
                r' DFT\| (.+? {2}) +(.+)', str_operation=str_to_header, repeats=True),
            Quantity('dft_u', r'(DFT\+U\|)'),
            Quantity('mp2', r'(MP2\|)'),
            Quantity('rpa', r'(RI-RPA\|)'),
            Quantity(
                'functional', r' FUNCTIONAL\| (\S+):', repeats=True),
            # TODO find example and verify
            Quantity(
                'vdw',
                r' vdW POTENTIAL\| +(.+? {2}) +(.+)', str_operation=str_to_header, repeats=True),
            Quantity(
                'qs',
                r' QS\| ((?:Method|Density cutoff)).*?:( {2}) +(.+)',
                str_operation=str_to_header, repeats=True),
            Quantity(
                'atomic_kind_information',
                r' ATOMIC KIND INFORMATION([\s\S]+?)\n\n\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'atom',
                    r'(ic kind: [A-Z][a-z]?[\s\S]+?)(?:\d+\. Atom|\Z)', repeats=True,
                    sub_parser=TextParser(quantities=[
                        Quantity('kind_label', r'ic kind:\s*(\w+)'),
                        Quantity('kind_number_of_atoms', r'Number of atoms:\s*(\d+)', dtype=int),
                        Quantity('kind_basis_set_name', r'Orbital Basis Set\s*(.+)'),
                        Quantity('basis_set_norm_type', r'Norm type:\s*(\d+)', dtype=int)
                    ] + n_orbital_basis_quantities))])),
            Quantity(
                'total_maximum_numbers',
                r' TOTAL NUMBERS AND MAXIMUM NUMBERS([\s\S]+?)\n\n\n',
                sub_parser=TextParser(quantities=[Quantity(
                    '%s' % key.lower().replace('the ', '').replace(' ', '_').replace('-', '_'),
                    r'\- %s:\s*(\d+)' % key, dtype=int) for key in [
                        'Atomic kinds', 'Atoms', 'Shell sets', 'Shells', 'Primitive Cartesian functions',
                        'Cartesian basis functions', 'Spherical basis functions',
                        'Orbital basis functions', 'Local part of the GTH pseudopotential',
                        'Non-local part of the GTH pseudopotential']])),
            Quantity(
                'atomic_coordinates',
                r' ATOMIC COORDINATES IN (angstrom[\s\S]+?)\n\n\n',
                convert=False, str_operation=str_to_atomic_coordinates),
            Quantity(
                'scf_parameters',
                r' SCF PARAMETERS([\s\S]+?)\*{79}',
                sub_parser=TextParser(quantities=[
                    Quantity('scf_max_iteration', r'max_scf:\s*(\d+)', dtype=int),
                    Quantity(
                        'scf_threshold_energy_change', rf'eps_scf:\s*({re_float})',
                        dtype=float, unit='hartree'),
                    Quantity(
                        'md',
                        r' MD\| (.+? {2}) +(.+)', str_operation=str_to_header, repeats=True
                    )])),
            # TODO add mp2, rpa, gw
            Quantity(
                'single_point',
                r'( Iteration\s*Convergence\s*Energy[\s\S]+?(?:\-{50}\n\s*\-|MD_ENERGIES))',
                sub_parser=TextParser(quantities=single_point_quantities)),
            Quantity(
                'geometry_optimization',
                r'STARTING GEOMETRY OPTIMIZATION([\s\S]+?(?:GEOMETRY OPTIMIZATION COMPLETED|\Z))',
                sub_parser=TextParser(quantities=geometry_optimization_quantities)),
            Quantity(
                'molecular_dynamics',
                r'(MD_ENERGIES\| Initialization proceeding[\s\S]+?\-{50}\n\s*\-)',
                sub_parser=TextParser(quantities=molecular_dynamics_quantities))
        ]

        self._quantities = [
            Quantity(
                'dbcsr',
                r' (DBCSR)\| (.+? {2}) +(.+)', str_operation=str_to_header, repeats=True),
            Quantity(
                'program',
                r'\*\*\s*PROGRAM ([\s\S]+?)(?:\*\*|\n\n|\Z)',
                str_operation=str_to_program, repeats=True),
            Quantity(
                'cp2k',
                r' CP2K\| (.+? {2}) +(.+)', str_operation=str_to_header, repeats=True),
            Quantity(
                'global',
                r' GLOBAL\| (.+? {2}) +(.+)', str_operation=str_to_header, repeats=True),
            Quantity(
                'restart',
                r'RESTART INFORMATION\s*\*+\s*\*+([\s\S]+?)\*{79}',
                sub_parser=TextParser(quantities=[
                    Quantity('filename', r'RESTART FILE NAME: (\S+)'),
                    Quantity(
                        'quantities',
                        r'RESTARTED QUANTITIES:\s*\*\s*([\s\S]+?)\Z',
                        str_operation=lambda x: [v.split('*')[1].strip() for v in x.strip().split('\n')])])),
            Quantity(
                'lattice_vectors',
                rf' CELL\| Vector [abc] \[angstrom\]:\s*({re_float})\s*({re_float})\s*({re_float})',
                repeats=True),
            # TODO add restart find example
            Quantity(
                'quickstep',
                r'\.\.\. make the atoms dance([\s\S]+?(?:\-{79}\s*\-|\Z))',
                sub_parser=TextParser(quantities=quickstep_quantities)),
            Quantity(
                'qs_dftb',
                r'  #####   #####        # ######  ####### ####### ######\s*'
                r' #     # #     #      #  #     # #          #    #     #\s*'
                r' #     # #           #   #     # #          #    #     #\s*'
                r' #     #  #####     #    #     # #####      #    ######\s*'
                r' #   # #       #   #     #     # #          #    #     #\s*'
                r' #    #  #     #  #      #     # #          #    #     #\s*'
                r'  #### #  #####  #       ######  #          #    ######\s*'
                r'([\s\S]+?(?:\-{79}\s*\-|\Z))',
                sub_parser=TextParser(quantities=quickstep_quantities))
            # TODO add other calculation types
        ]


class CP2KParser(FairdiParser):
    def __init__(self):
        super().__init__(
            name='parsers/cp2k', code_name='CP2K', code_homepage='https://www.cp2k.org/',
            mainfile_contents_re=(
                r'\*\*\*\* \*\*\*\* \*\*\*\*\*\*  \*\*  PROGRAM STARTED AT\s.*\n'
                r' \*\*\*\*\* \*\* \*\*\*  \*\*\* \*\*   PROGRAM STARTED ON\s*.*\n'
                r' \*\*    \*\*\*\*   \*\*\*\*\*\*    PROGRAM STARTED BY .*\n'
                r' \*\*\*\*\* \*\*    \*\* \*\* \*\*   PROGRAM PROCESS ID .*\n'
                r'  \*\*\*\* \*\*  \*\*\*\*\*\*\*  \*\*  PROGRAM STARTED IN .*\n')
        )
        self._metainfo_env = m_env
        self.out_parser = CP2KOutParser()
        self.inp_parser = InpParser()
        # use a custom xyz parser as the output of cp2k is sometimes not up to standard
        self.traj_parser = TrajParser(type='positions')
        self.velocities_parser = TrajParser(type='velocities')
        self.cell_parser = DataTextParser()
        self.energy_parser = DataTextParser()
        self.force_parser = ForceParser()
        self._method = None
        self._calculation_type = None

        # TODO add vdw parameter
        self._metainfo_name_map = {
            'started_at': 'start_time', 'started_on': 'start_host',
            'started_by': 'start_user', 'process_id': 'id', 'started_in': 'start_path',
            'ended_at': 'end_time', 'ran_on': 'end_host', 'ran_by': 'end_user',
            'stopped_in': 'end_path', 'version_string:': 'program_version',
            'source_code_revision_number:': 'svn_revision',
            'program_compiled_at': 'program_compilation_datetime',
            'program_compiled_on': 'program_compilation_host',
            'input_file_name': 'input_filename',
            'basis_set_file_name': 'basis_set_filename',
            'geminal_file_name': 'geminal_filename',
            'potential_file_name': 'potential_filename',
            'mm_potential_file_name': 'mm_potential_filename',
            'coordinate_file_name': 'coordinate_filename',
            'preferred_diagonalization_lib.': 'preferred_diagonalization_library',
            'spin_restricted_kohn-sham_(rks)_calculation': 'spin_restriction',
            'multiplicity': 'spin_target_multiplicity',
            'number_of_spin_states': 'number_of_spin_channels', 'charge': 'total_charge',
            'self-interaction_correction_(sic)': 'self_interaction_correction_method',
            'method': 'quickstep_method', 'density_cutoff': 'planewave_cutoff',
            'temperature': 'target_temperature', 'temperature_tolerance': 'target_temperature_tolerance',
            'pressure': 'target_pressure', 'print_md_information_every': 'print_frequency',
            'potential_form:': 'vdw_name', 'bj_damping:': 'bj_damping_name',
            'cutoff_radius_[bohr]:': 'cutoff_radius', 'scaling_factor:': 'scaling_factor',
            'exp_prefactor_for_damping:': 'damping_factor', 's6_scaling_factor:': 's6_scaling_factor',
            'sr6_scaling_factor:': 'sr6_scaling_factor', 's8_scaling_factor:': 's6_scaling_factor',
            'cutoff_for_cn_calculation:': 'cn_cutoff', 'optimization_method': 'method',
            'total_energy': 'energy', 'real_energy_change': 'energy_change',
            'decrease_in_energy': 'energy_decrease', 'conv_limit_for_step_size': 'step_size_convergence_limit',
            'convergence_in_step_size': 'step_size_convergence',
            'convergence_in_rms_step_size': 'rms_step_size_convergence',
            'conv_limit_for_gradients': 'gradient_convergence_limit',
            'conv_for_gradients': 'max_gradient_convergence',
            'conv_in_rms_gradients': 'rms_gradient_convergence',
            'exchange_correlation_energy': 'energy_XC'}

        self._self_interaction_map = {
            'NO': None, 'D SIC': 'SIC_AD', 'Explicit Orbital SIC': 'SIC_EXPLICIT_ORBITALS',
            'SPZ/MAURI SIC': 'SIC_MAURI_SPZ', 'US/MAURI SIC': 'SIC_MAURI_US'}
        self._optimization_method_map = {
            'CONJUGATE GRADIENTS': 'conjugate gradient', 'BFGS': 'bfgs', 'L-BFGS': 'bfgs'}
        self._file_extension_map = {
            'XYZ': 'xyz', 'XMOL': 'xyz', 'ATOMIC': 'xyz', 'PDB': 'pdb', 'DCD': 'dcd'}
        self._xc_functional_map = {
            'BLYP': [XCFunctional('GGA_X_B88'), XCFunctional('GGA_C_LYP')],
            'LDA': [XCFunctional('LDA_XC_TETER93')],
            'PADE': [XCFunctional('LDA_XC_TETER93')],
            'PBE': [XCFunctional('GGA_X_PBE'), XCFunctional('GGA_C_PBE')],
            'OLYP': [XCFunctional('GGA_X_OPTX'), XCFunctional('GGA_C_LYP')],
            'HCTH120': [XCFunctional('GGA_XC_HCTH_120')],
            'PBE0': [XCFunctional('HYB_GGA_XC_PBEH')],
            'B3LYP': [XCFunctional('HYB_GGA_XC_B3LYP')],
            'TPSS': [XCFunctional('MGGA_X_TPSS'), XCFunctional('MGGA_C_TPSS')]}
        self._ensemble_map = {'NVE': 'NVE', 'NVT': 'NVT', 'NPT_F': 'NPT', 'NPT_I': 'NPT'}
        self._vdw_map = {
            "S. Grimme, JCC 27: 1787 (2006)": "G06",
            "S. Grimme et al, JCP 132: 154104 (2010)": "G10"}

        self._settings = None

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self.inp_parser.mainfile = None
        self.traj_parser.mainfile = None
        self.velocities_parser.mainfile = None
        self.energy_parser.mainfile = None
        self.force_parser.mainfile = None
        self.out_parser.logger = self.logger
        self.inp_parser.logger = self.logger
        self.traj_parser.logger = self.logger
        self.velocities_parser.logger = self.logger
        self.energy_parser.logger = self.logger
        self.force_parser.logger = self.logger
        self._settings = None
        self._method = None
        self._calculation_type = None

    @property
    def settings(self):
        if self._settings is None:
            def to_dict(data):
                data_dict = dict()
                for key, val in data:
                    name = self._metainfo_name_map.get(key, key)
                    data_dict.setdefault(name, [])
                    data_dict[name].append(val)
                for key, val in data_dict.items():
                    data_dict[key] = val[0] if len(val) == 1 else val
                return data_dict

            self._settings = dict()
            self._settings['dft'] = to_dict(
                self.out_parser.get(self._calculation_type, {}).get('dft', []))
            self._settings['qs'] = to_dict(
                self.out_parser.get(self._calculation_type, {}).get('qs', []))
            self._settings['vdw'] = to_dict(
                self.out_parser.get(self._calculation_type, {}).get('vdw', []))
            self._settings['dbcsr'] = to_dict(self.out_parser.get('dbcsr', []))
            self._settings['program'] = to_dict(self.out_parser.get('program', []))
            self._settings['cp2k'] = to_dict(self.out_parser.get('cp2k', []))
            self._settings['global'] = to_dict(self.out_parser.get('global', []))
            self._settings['md'] = to_dict(
                self.out_parser.get(self._calculation_type, {}).get('scf_parameters', {}).get('md', []))

        return self._settings

    def _normalize_filename(self, filename):
        if filename.startswith('='):
            filename = filename[1:]
        elif re.match(r'./', filename):
            filename = filename
        else:
            project_name = self.inp_parser.get('GLOBAL/PROJECT_NAME')
            if filename:
                filename = '%s-%s' % (project_name, filename)
            else:
                filename = project_name
        return filename

    def get_atomic_number(self, element):
        atomic_numbers = self.out_parser.get(
            self._calculation_type, {}).get('atomic_coordinates').atomic_numbers
        return atomic_numbers.get(element, 0)

    def get_ensemble_type(self, frame):
        if self.sampling_method != 'molecular_dynamics':
            return

        if frame == 0:
            return self.settings['md'].get('ensemble_type', '')
        else:
            calculation = self.out_parser.get(self._calculation_type, '')
            if not calculation:
                return calculation
            return calculation.molecular_dynamics.md_step[frame - 1].get('ensemble_type', '')

    def get_velocities(self, frame):
        if self.out_parser.get(self._calculation_type, {}).get('molecular_dynamics') is not None:
            return

        if self.velocities_parser.mainfile is None:
            frequency, filename = self.settings['md'].get('velocities', '0 none').split()
            frequency = int(frequency)
            if frequency == 0:
                filename = '%s-vel-1.xyz' % self.inp_parser.get('GLOBAL/PROJECT_NAME')
                frequency = 1

            self.velocities_parser.mainfile = os.path.join(self.maindir, filename)
            self.velocities_parser.units = resolve_unit(
                self.inp_parser.get('MOTION/PRINT/VELOCITIES/UNIT', 'bohr*au_t^-1'))
            self.velocities_parser._frequency = frequency

        if self.get_ensemble_type(frame).lower() == 'REFTRAJ':
            frame -= 1

        if frame % self.velocities_parser._frequency != 0 or frame < 0:
            return

        try:
            return self.velocities_parser.trajectory[frame // self.velocities_parser._frequency]
        except Exception:
            self.logger.error('Error reading velocities.')

    def get_trajectory(self, frame):
        trajectory = None

        if frame == 0:
            coord = self.inp_parser.get('FORCE_EVAL/SUBSYS/COORD/DEFAULT_KEYWORD')
            units = resolve_unit(self.inp_parser.get('FORCE_EVAL/SUBSYS/COORD/UNIT', 'angstrom'))
            if coord is None:
                coord_filename = self.inp_parser.get('FORCE_EVAL/SUBSYS/TOPOLOGY/COORD_FILE_NAME', '')
                self.traj_parser.mainfile = os.path.join(self.maindir, coord_filename.strip())
                self.traj_parser.units = units
                if self.traj_parser.trajectory is not None:
                    result = self.traj_parser.trajectory[0]
                    # reset for output trajectory
                    self.traj_parser.mainfile = None
                    trajectory = result

            else:
                coord = np.transpose([c.split() for c in coord])
                positions = np.array(coord[1:4], dtype=float).T * units
                scaled = 'T' in self.inp_parser.get('FORCE_EVAL/SUBSYS/COORD/SCALED', 'False')
                if scaled:
                    trajectory = Trajectory(labels=coord[0], scaled_positions=positions)
                else:
                    trajectory = Trajectory(labels=coord[0], positions=positions)

        if trajectory is not None:
            return trajectory

        if self.traj_parser.mainfile is None:
            # try to get it from md
            frequency, filename = self.settings['md'].get('coordinates', '0 none').split()
            frequency = int(frequency)
            if frequency == 0:
                filename = self.inp_parser.get('MOTION/PRINT/TRAJECTORY/FILENAME', '').strip()
                filename = self._normalize_filename(filename)
                traj_format = self.inp_parser.get('MOTION/PRINT/TRAJECTORY/FORMAT', 'XYZ').strip()
                traj_format = self._file_extension_map.get(traj_format, 'xyz')
                filename = '%s-pos-1.%s' % (filename, traj_format)
                frequency = 1

            self.traj_parser.mainfile = os.path.join(self.maindir, filename)
            self.traj_parser.units = resolve_unit(
                self.inp_parser.get('MOTION/PRINT/TRAJECTORY/UNIT', 'angstrom'))
            self.traj_parser._frequency = frequency

        if self.get_ensemble_type(frame) == 'REFTRAJ':
            frame -= 1

        if frame % self.traj_parser._frequency != 0 or frame < 0:
            return

        try:
            return self.traj_parser.trajectory[frame // self.traj_parser._frequency]
        except Exception:
            self.logger.error('Error reading trajectory.')

    def get_lattice_vectors(self, frame):
        lattice_vectors = None

        if frame == 0:
            lattice_vectors = self.out_parser.get('lattice_vectors')
            if lattice_vectors is None:
                # get it from input
                cell = self.inp_parser.get('FORCE_EVAL/SUBSYS/CELL')
                # is this the unit for cell? how about for angles
                units = resolve_unit(
                    self.inp_parser.get('FORCE_EVAL/SUBSYS/COORD/UNIT', 'angstrom'))
                if cell is None:
                    return

                if 'A' in cell and 'B' in cell and 'C' in cell:
                    lattice_vectors = np.array([
                        cell.get(c).split() for c in ('A', 'B', 'C')], dtype=float) * units
                elif 'ABC' in cell:
                    abc = (np.array(
                        cell.get('ABC').split(), dtype=float) * units).to('angstrom').magnitude
                    angles = np.array(cell.get('ALPHA_BETA_GAMMA', '90. 90. 90.').split(), dtype=float)
                    lattice_vectors = ase.geometry.cellpar_to_cell(np.hstack((abc, angles))) * ureg.angstrom

            else:
                units = resolve_unit(
                    self.inp_parser.get('FORCE_EVAL/SUBSYS/COORD/UNIT', 'angstrom'))
                lattice_vectors = lattice_vectors * units

        if lattice_vectors is not None:
            return lattice_vectors

        if self.cell_parser.mainfile is None:
            frequency, filename = self.settings['md'].get('simulation_cell', '0 none').split()
            frequency = int(frequency)
            if frequency == 0:
                # TODO test this I cannot find a sample output cell filee
                filename = self.inp_parser.get('MOTION/PRINT/CELL/FILENAME', '').strip()
                frequency = 1

            if filename:
                self.cell_parser.mainfile = os.path.join(self.maindir, filename)
                self.cell_parser.units = resolve_unit(
                    self.inp_parser.get('MOTION/PRINT/TRAJECTORY/UNIT', 'angstrom'))
                self.cell_parser._frequency = frequency
            else:
                if self.sampling_method == 'molecular_dynamics':
                    # check that this is not an NPT
                    ensemble_type = self.get_ensemble_type(frame)
                    if ensemble_type[:3] == 'NPT':
                        return
                return self.get_lattice_vectors(0)

        if self.get_ensemble_type(frame) == 'REFTRAJ':
            frame -= 1

        if frame % self.cell_parser._frequency != 0 or frame < 0:
            return

        try:
            return self.cell_parser.data[frame // self.cell_parser._frequency] * resolve_unit(self.cell_parser.units)
        except Exception:
            self.logger.error('Error reading lattice vectors.')

    def get_md_output(self, frame):
        if self.energy_parser.mainfile is None:
            frequency, filename = self.settings['md'].get('energies', '0, none').split()
            frequency = int(frequency)
            if frequency == 0:
                return
            self.energy_parser.mainfile = os.path.join(self.maindir, filename)
            self.energy_parser._frequency = frequency

        if self.get_ensemble_type(frame) == 'REFTRAJ':
            frame -= 1

        if frame % self.energy_parser._frequency != 0 or frame < 0:
            return dict()

        try:
            data = self.energy_parser.data[frame // self.energy_parser._frequency]
            return dict(
                time=data[1],
                kinetic_energy_instantaneous=data[2] * ureg.hartree,
                temperature_instantaneous=data[3],
                potential_energy_instantaneous=data[4] * ureg.hartree,
                conserved_quantity=data[5] * ureg.hartree,
                cpu_time_instantaneous=data[6])

        except Exception:
            self.logger.error('Error reading MD energies.')

    def get_forces(self, frame):
        filename = self.inp_parser.get('FORCE_EVAL/PRINT/FORCES/FILENAME', '').strip()
        filename = self._normalize_filename(filename)
        filename = '%s-1_%d.xyz' % (filename, frame)
        self.force_parser.mainfile = os.path.join(self.maindir, filename)
        return self.force_parser.get('atom_forces')

    def get_xc_functionals(self):
        functionals = self.inp_parser.get('FORCE_EVAL/DFT/XC/XC_FUNCTIONAL/VALUE')
        if functionals is None or functionals == 'NO_SHORTCUT':
            functional_values = self.inp_parser.get('FORCE_EVAL/DFT/XC/XC_FUNCTIONAL', {})
            functionals = []
            for name, attrib in functional_values.items():
                name = name.upper()
                if name == 'VALUE':
                    continue
                # get xc_func from mapping then apply read attributes
                # if func is not in mapping, create it
                values = self._xc_functional_map.get(name, [XCFunctional(name)])
                for n, value in enumerate(values):
                    weight = attrib.get('SCALE_X', None) if n == 0 else attrib.get('SCALE_C', None)
                    value._data.update({'weight': weight})
                functionals.extend(values)
        else:
            names = [functionals] if not isinstance(functionals, list) else functionals
            functionals = []
            for name in names:
                name = name.upper()
                if name not in self._xc_functional_map:
                    self.logger.error('Cannot resolve xc functional')
                    continue
                functionals.extend(self._xc_functional_map.get(name))

        return functionals

    def parse_scc(self, source):
        if source is None:
            return

        sec_scc = self.archive.section_run[-1].m_create(SingleConfigurationCalculation)

        for key in ['energy_total', 'stress_tensor']:
            val = source.get(key)
            if val is not None:
                setattr(sec_scc, key, val)

        for key in ['electronic_kinetic_energy', 'exchange_correlation_energy']:
            val = source.get(key)
            if val is not None:
                key = self._metainfo_name_map.get(key, key)
                setattr(sec_scc, key, val[-1])

        # self consistency
        for iteration in source.get('iteration', []):
            sec_scf = sec_scc.m_create(ScfIteration)
            for key, val in iteration.items():
                if val is not None:
                    setattr(sec_scf, key, val)

        atom_forces = source.get('atom_forces', self.get_forces(source._frame))
        if atom_forces is not None:
            atom_forces = np.array(atom_forces) * ureg.hartree / ureg.bohr
            sec_scc.atom_forces = atom_forces

        # TODO add dos
        return sec_scc

    def parse_system(self, trajectory):

        trajectory = 0 if trajectory is None else trajectory

        if isinstance(trajectory, int):
            frame = trajectory
            trajectory = self.get_trajectory(frame)
            if trajectory is not None:
                trajectory._frame = frame

        if trajectory is None:
            return

        sec_system = self.archive.section_run[-1].m_create(System)

        lattice_vectors = self.get_lattice_vectors(trajectory._frame)

        if trajectory.positions is not None:
            sec_system.atom_positions = trajectory.positions
        elif trajectory.scaled_positions is not None and lattice_vectors is not None:
            sec_system.atom_positions = np.dot(trajectory.scaled_positions, lattice_vectors)

        labels = trajectory.labels if trajectory.labels is not None else self.out_parser.get(
            self._calculation_type).get('atomic_coordinates')
        if labels is not None:
            sec_system.atom_labels = labels

        if lattice_vectors is not None:
            sec_system.lattice_vectors = lattice_vectors
        periodic = self.inp_parser.get('FORCE_EVAL/SUBSYS/CELL/PERIODIC', 'xyz').lower()
        sec_system.configuration_periodic_dimensions = [v in periodic for v in ('x', 'y', 'z')]

        # TODO test this I cannot find an example
        # velocities
        if self.sampling_method == 'molecular_dynamics':
            velocities = self.get_velocities(trajectory._frame)
            if velocities is not None:
                sec_system.atom_velocities = velocities

        return sec_system

    def parse_configurations_quickstep(self):
        sec_run = self.archive.section_run[-1]
        quickstep = self.out_parser.get(self._calculation_type)

        # quickstep extension to scc quantities
        def parse_quickstep_calculation(source):
            if source is None:
                return
            sec_quickstep_calc = sec_run.m_create(x_cp2k_section_quickstep_calculation)
            for key in ['energy_total', 'atom_forces']:
                val = source.get(key)
                if val is not None:
                    setattr(sec_quickstep_calc, 'x_cp2k_%s' % key, val)
                sec_quickstep_calc.x_cp2k_quickstep_converged = source.get('converged') is not None

            if source.get('electronic_kinetic_energy') is not None:
                sec_quickstep_calc.x_cp2k_electronic_kinetic_energy = source.get('electronic_kinetic_energy')[-1]

            for iteration in source.get('iteration', []):
                sec_scf = sec_quickstep_calc.m_create(x_cp2k_section_scf_iteration)
                for key, val in iteration.items():
                    if val is not None:
                        setattr(sec_scf, 'x_cp2k_%s' % key, val)

            if source.stress_tensor is not None:
                sec_stress = sec_quickstep_calc.m_create(x_cp2k_section_stress_tensor)
                if source.stress_tensor_one_third_of_trace is not None:
                    val = source.stress_tensor_one_third_of_trace
                    sec_stress.stress_tensor_one_third_of_trace = val.to('pascal').magnitude
                if source.stress_tensor_determinant is not None:
                    sec_stress.stress_tensor_determinant = val.to('pascal**3').magnitude
                if source.stress_eigenvalues_eigenvectors is not None:
                    val = source.stress_eigenvalues_eigenvectors
                    sec_stress.x_cp2k_stress_tensor_eigenvalues = val[0].to('pascal').magnitude
                    sec_stress.x_cp2k_stress_tensor_eigenvectors = val[1]

        def parse_md_step(source):
            # we put md output in scc, originally in frame sequence
            # TODO put in workflow
            md_output = self.get_md_output(source._frame)
            md_output = md_output if md_output else source
            sec_scc = sec_run.section_single_configuration_calculation[-1]
            sec_md_step = sec_scc.m_create(x_cp2k_section_md_step)

            with_average = [
                'cpu_time', 'energy_drift', 'potential_energy', 'kinetic_energy',
                'temperature', 'pressure', 'barostat_temperature', 'volume']
            for key, val in md_output.items():
                if val is None:
                    continue
                name = 'x_cp2k_md_%s' % key

                if key == 'energy_drift':
                    pass
                elif 'energy' in key or 'conserved' in key:
                    val = val.to('joule').magnitude
                elif 'pressure' in key:
                    val - val.to('pascal').magnitude
                elif 'volume' in key:
                    val = val.to('m**3').magnitude

                if key in with_average:
                    setattr(sec_md_step, '%s_instantaneous' % name, val[0])
                    name = '%s_average' % name
                    val = val[1]

                setattr(sec_md_step, name, val)

        def parse_calculations(calculations):
            for n, calculation in enumerate(calculations):
                self_consistent = calculation.get('self_consistent', [])
                self_consistent = [self_consistent] if not isinstance(self_consistent, list) else self_consistent
                # there may be several wave function optimizations in a calculation
                for scf in self_consistent:
                    parse_quickstep_calculation(scf)

                # write only the last one to scc
                scf = self_consistent[-1] if self_consistent else calculation
                scf._frame = n
                sec_scc = self.parse_scc(scf)
                if calculation.get('ensemble_type') is not None:
                    calculation._frame = n
                    parse_md_step(calculation)

                if n == 0:
                    atomic_coord = quickstep.get('atomic_coordinates')
                    if atomic_coord is not None:
                        atomic_coord._frame = 0
                    sec_system = self.parse_system(atomic_coord)
                else:
                    sec_system = self.parse_system(n)

                if sec_system is not None:
                    sec_scc.single_configuration_calculation_to_system_ref = sec_system

            sec_scc.single_configuration_to_calculation_method_ref = sec_run.section_method[-1]

        single_point = quickstep.get('single_point')
        if single_point is not None:
            parse_calculations([single_point])

        geometry_optimization = quickstep.get('geometry_optimization')
        if geometry_optimization is not None:
            # initial self consistent
            optimization_steps = [geometry_optimization]
            optimization_steps.extend(geometry_optimization.get('optimization_step', []))
            parse_calculations(optimization_steps)

        molecular_dynamics = quickstep.get('molecular_dynamics')
        if molecular_dynamics is not None:
            md_steps = [molecular_dynamics]
            md_steps.extend(molecular_dynamics.get('md_step', []))
            parse_calculations(md_steps)

    def parse_method_quickstep(self):
        sec_run = self.archive.section_run[-1]
        sec_method = sec_run.m_create(Method)
        quickstep = self.out_parser.get(self._calculation_type)

        # electronic structure method
        if quickstep.get('dft') is not None:
            sec_method.electronic_structure_method = 'DFT'
        elif quickstep.get('dft_u') is not None:
            sec_method.electronic_structure_method = 'DFT+U'
        elif quickstep.get('mp2') is not None:
            sec_method.electronic_structure_method = 'MP2'
        elif quickstep.get('rpa') is not None:
            sec_method.electronic_structure_method = 'RPA'

        # xc functionals
        for functional in self.get_xc_functionals():
            sec_xc_functional = sec_method.m_create(XCFunctionals)
            sec_xc_functional.XC_functional_name = functional.name
            sec_xc_functional.XC_functional_weight = functional.weight

        # van der Waals settings
        # TODO test this no example, add parameter
        vdw = self.settings['vdw']
        if vdw:
            sec_vdw = sec_method.m_create(x_cp2k_section_vdw_settings)
            for key, val in vdw.items():
                if val == '':
                    setattr(sec_vdw, 'x_cp2k_vdw_name', key)
                    vdw_name = self._vdw_map.get(key, None)
                    if vdw_name is not None:
                        sec_method.van_der_Waals_method = vdw_name
                else:
                    setattr(sec_vdw, 'x_cp2k_vdw_%s' % key, val)

        stress_method = self.inp_parser.get('FORCE_EVAL/STRESS_TENSOR')
        if stress_method is not None:
            sec_method.stress_tensor_method = stress_method.replace('_', ' ').title()

        sec_quickstep_settings = sec_method.m_create(x_cp2k_section_quickstep_settings)
        if self.settings['dft']:
            for key, val in self.settings['dft'].items():
                section = sec_method
                if key == 'self_interaction_correction_method':
                    val = self._self_interaction_map.get(val, None)
                elif key == 'spin_restriction':
                    section = sec_quickstep_settings
                    key = 'x_cp2k_%s' % key
                if val is None:
                    continue
                setattr(section, key, val)
        if self.settings['qs']:
            for key, val in self.settings['qs'].items():
                setattr(sec_quickstep_settings, 'x_cp2k_%s' % key, val)

        atomic_kind_info = quickstep.get('atomic_kind_information', None)
        if atomic_kind_info is not None:
            sec_atom_kinds = sec_quickstep_settings.m_create(x_cp2k_section_atomic_kinds)
            for atom in atomic_kind_info.get('atom', []):
                # why necessary to make a separate section
                sec_atom_kind = sec_atom_kinds.m_create(x_cp2k_section_atomic_kind)
                sec_kind_basis_set = sec_atom_kind.m_create(x_cp2k_section_kind_basis_set)
                for key, val in atom.items():
                    if val is None:
                        continue
                    if key in ['kind_label', 'kind_number_of_atoms']:
                        setattr(sec_atom_kind, 'x_cp2k_%s' % key, str(val))
                    else:
                        setattr(sec_kind_basis_set, 'x_cp2k_%s' % key, val)

                sec_method_atom_kind = sec_method.m_create(MethodAtomKind)
                sec_method_atom_kind.method_atom_kind_label = atom.kind_label
                sec_method_atom_kind.method_atom_kind_atom_number = self.get_atomic_number(atom.kind_label)

        total_maximum_numbers = quickstep.get('total_maximum_numbers', None)
        if total_maximum_numbers is not None:
            sec_total = sec_quickstep_settings.m_create(x_cp2k_section_total_numbers)
            sec_maximum = sec_quickstep_settings.m_create(x_cp2k_section_maximum_angular_momentum)
            for key, val in total_maximum_numbers.items():
                if val is None:
                    continue
                if key in ['orbital_basis_functions', 'local_part_of_gth_pseudopotential', 'non_local_part_of_gth_pseudopotential']:
                    setattr(sec_maximum, 'x_cp2k_%s' % key, val)
                else:
                    setattr(sec_total, 'x_cp2k_%s' % key, val)

        scf_parameters = quickstep.get('scf_parameters', None)
        if scf_parameters is not None:
            for key, val in scf_parameters.items():
                if val is None:
                    continue
                setattr(sec_method, key, val)

        sec_method_basis_set = sec_method.m_create(MethodBasisSet)
        sec_method_basis_set.method_basis_set_kind = 'wavefunction'
        sec_method_basis_set.number_of_basis_sets_atom_centered = len(sec_run.section_basis_set_atom_centered)

    @property
    def sampling_method(self):
        if self._method is None:
            quickstep = self.out_parser.get(self._calculation_type, {})
            for method in ['single_point', 'geometry_optimization', 'molecular_dynamics']:
                if quickstep.get(method) is not None:
                    self._method = method
        return self._method

    def parse_sampling_method(self):
        # TODO move these all to workflow
        # TODO add vdW
        sec_sampling_method = self.archive.section_run[0].m_create(SamplingMethod)
        sec_sampling_method.sampling_method = self.sampling_method

        if self.sampling_method == 'geometry_optimization':
            optimization = self.out_parser.get(self._calculation_type).geometry_optimization
            if optimization.method is not None:
                method = self._optimization_method_map.get(optimization.method, '')
                if not method:
                    self.logger.error('Cannot resolve optimization method.')
                sec_sampling_method.geometry_optimization_method = method
            sec_geometry_opt = sec_sampling_method.m_create(x_cp2k_section_geometry_optimization)
            for step in optimization.get('optimization_step', []):
                information = step.information
                if information is None:
                    continue
                sec_geometry_opt_step = sec_geometry_opt.m_create(x_cp2k_section_geometry_optimization_step)
                for key, val in information.items():
                    if val is None:
                        continue

                    name = self._metainfo_name_map.get(key, key)
                    if name.startswith('energy') and isinstance(val, float):
                        val = (val * ureg.hartree).to('joule').magnitude
                    elif 'step_size' in name and isinstance(val, float):
                        val = (val * ureg.bohr).to('m').magnitude
                    elif 'gradient' in name and isinstance(val, float):
                        val = (val * ureg.hartree / ureg.bohr).to('joule/m').magnitude
                    elif isinstance(val, str):
                        val = val.strip()

                    setattr(sec_geometry_opt_step, 'x_cp2k_optimization_%s' % name, val)

            geometry_change = sec_geometry_opt_step.x_cp2k_optimization_step_size_convergence_limit
            if geometry_change is not None:
                sec_sampling_method.geometry_optimization_geometry_change = geometry_change
            threshold_force = sec_geometry_opt_step.x_cp2k_optimization_gradient_convergence_limit
            if threshold_force is not None:
                sec_sampling_method.geometry_optimization_threshold_force = threshold_force

        elif self.sampling_method == 'molecular_dynamics':
            ensemble_type = self._ensemble_map.get(self.get_ensemble_type(0), None)
            if ensemble_type is not None:
                sec_sampling_method.ensemble_type = ensemble_type

            sec_md_settings = sec_sampling_method.m_create(x_cp2k_section_md_settings)
            for key, val in self.settings['md'].items():
                if val is None or key == 'file_type':
                    continue
                if key in ['coordinates', 'simulation_cell', 'velocities', 'energies', 'dump']:
                    val = val.split()
                    setattr(sec_md_settings, 'x_cp2k_md_%s_print_frequency' % key, int(val[0]))
                    setattr(sec_md_settings, 'x_cp2k_md_%s_filename' % key, val[1])
                elif key == 'print_frequency':
                    setattr(sec_md_settings, 'x_cp2k_md_%s' % key, int(val.split()[0]))
                else:
                    setattr(sec_md_settings, 'x_cp2k_md_%s' % key, val)

    def parse_input(self):
        # TODO include extended input
        input_filename = self.settings['cp2k'].get('input_filename', None)
        if input_filename is None:
            return

        def resolve_definition(name):
            return self._metainfo_env.all_definitions_by_name.get(name, [None])[0]

        def override_keyword(name):
            # override keys to be compatible with metainfo name
            # TODO change metainfo name
            if name.endswith('_VALUE'):
                return name.replace('VALUE', 'SECTION_PARAMETERS')
            elif name.endswith('KIND_RI_AUX_BASIS'):
                return name.replace('BASIS', 'BASIS_SET')
            return name

        def parse(name, data, section):
            if isinstance(data, InpValue):
                sec_def = resolve_definition(name)
                if sec_def is not None:
                    sub_section = section.m_create(sec_def.section_cls)
                    for key, val in data.items():
                        sec_name = '%s_%s' % (name, key)
                        parse(sec_name, val, sub_section)

            elif isinstance(data, list) and data:
                for val in data:
                    parse(name, val, section)

            else:
                name = name.replace('_section', '')
                name = override_keyword(name)
                quantity_def = resolve_definition(name)
                if quantity_def is not None:
                    setattr(section, name, quantity_def.type(data))

        self.inp_parser.mainfile = os.path.join(self.maindir, input_filename)
        if self.inp_parser.tree is None:
            return

        parse('x_cp2k_section_input', self.inp_parser.tree, self.archive.section_run[-1])

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging.getLogger(__name__)

        self.init_parser()

        # identify calculation type, TODO add more
        calculation_types = ['quickstep', 'qs_dftb']
        for calculation_type in calculation_types:
            if self.out_parser.get(calculation_type) is not None:
                self._calculation_type = calculation_type
                break

        sec_run = self.archive.m_create(Run)
        sec_run.program_name = 'CP2K'
        sec_run.program_basis_set_type = 'gaussians'
        sec_run.program_version = self.settings['cp2k']['program_version']
        sec_run.program_compilation_host = self.settings['cp2k']['program_compilation_host']

        if self.settings['dbcsr']:
            sec_dbcsr = sec_run.m_create(x_cp2k_section_dbcsr)
            for key, val in self.settings['dbcsr'].items():
                setattr(sec_dbcsr, 'x_cp2k_%s' % key, val)

        if self.settings['program']:
            sec_startinformation = sec_run.m_create(x_cp2k_section_startinformation)
            sec_endinformation = sec_run.m_create(x_cp2k_section_end_information)
            section = sec_startinformation
            for key, val in self.settings['program'].items():
                if key == 'id':
                    sec_endinformation.x_cp2k_end_id = val[1]
                    key, val = 'start_id', val[0]
                section = sec_endinformation if key.startswith('end') else sec_startinformation
                setattr(section, 'x_cp2k_%s' % key, val)

        if self.settings['cp2k']:
            sec_program_information = sec_run.m_create(x_cp2k_section_program_information)
            for key, val in self.settings['cp2k'].items():
                if key == 'svn_revision':
                    try:
                        val = int(val.strip('svn:'))
                    except Exception:
                        continue
                setattr(sec_program_information, 'x_cp2k_%s' % key, val)

        if self.settings['global']:
            sec_global_settings = sec_run.m_create(x_cp2k_section_global_settings)
            for key, val in self.settings['global'].items():
                setattr(sec_global_settings, 'x_cp2k_%s' % key, val)

        planewave_cutoff = self.settings.get('qs', {}).get('planewave_cutoff', None)
        if planewave_cutoff is not None:
            sec_basis_set = sec_run.m_create(BasisSetCellDependent)
            sec_basis_set.basis_set_planewave_cutoff = planewave_cutoff * ureg.hartree

        atoms = self.out_parser.get(
            self._calculation_type, {}).get('atomic_kind_information', {}).get('atom', [])
        for atom in atoms:
            basis_set = atom.get('kind_basis_set_name', None)
            if basis_set is not None:
                sec_basis_set = sec_run.m_create(BasisSetAtomCentered)
                sec_basis_set.basis_set_atom_number = self.get_atomic_number(atom.kind_label)
                sec_basis_set.basis_set_atom_centered_short_name = basis_set

        restart = self.out_parser.get('restart')
        if restart is not None:
            sec_restart = sec_run.m_create(x_cp2k_section_restart_information)
            sec_restart.x_cp2k_restart_file_name = restart.get('filename')
            sec_restart.x_cp2k_restarted_quantity_name = ' '.join(restart.get('quantities'))

        self.parse_input()

        if self._calculation_type in ['quickstep', 'qs_dftb']:
            self.parse_method_quickstep()
            self.parse_configurations_quickstep()

        self.parse_sampling_method()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import os
import time
import click
import urllib.parse
import requests

from nomad import config
from nomad import processing

from .client import client


def stream_upload_with_client(client, stream, name=None):
    user = client.auth.get_auth().response().result
    token = user.access_token
    url = config.client.url + '/uploads/'
    if name is not None:
        url += '?name=%s' % urllib.parse.quote(name)

    response = requests.put(url, headers={'Authorization': 'Bearer %s' % token}, data=stream)
    if response.status_code != 200:
        raise Exception('nomad return status %d' % response.status_code)
    upload_id = response.json()['upload_id']

    return client.uploads.get_upload(upload_id=upload_id).response().result


def upload_file(file_path: str, name: str = None, offline: bool = False, publish: bool = False, client=None):
    '''
    Upload a file to nomad.

    Arguments:
        file_path: path to the file, absolute or relative to call directory
        name: optional name, default is the file_path's basename
        offline: allows to process data without upload, requires client to be run on the server
        publish: automatically publish after successful processing

    Returns: The upload_id
    '''
    if client is None:
        from nomad.cli.client import create_client
        client = create_client()
    if offline:
        upload = client.uploads.upload(
            local_path=os.path.abspath(file_path), name=name).response().result
        click.echo('process offline: %s' % file_path)
    else:
        # bravado does not seem to support streaming?
        try:
            with open(file_path, 'rb') as f:
                upload = stream_upload_with_client(client, f, name=name)
        except Exception as e:
            click.echo('could not upload the file: %s' % str(e))
            return

    while upload is not None and upload.tasks_status not in [processing.SUCCESS, processing.FAILURE]:
        upload = client.uploads.get_upload(upload_id=upload.upload_id).response().result
        calcs = upload.calcs.pagination
        if calcs is None:
            total, successes, failures = 0, 0, 0
        else:
            total, successes, failures = (calcs.total, calcs.successes, calcs.failures)

        ret = '\n' if upload.tasks_status in (processing.SUCCESS, processing.FAILURE) else '\r'

        print(
            'status: %s; task: %s; parsing: %d/%d/%d                %s' %
            (upload.tasks_status, upload.current_task, successes, failures, total, ret), end='')

        time.sleep(1)

    if upload.tasks_status == processing.FAILURE:
        click.echo('There have been errors:')
        for error in upload.errors:
            click.echo('    %s' % error)
    elif publish:
        client.uploads.exec_upload_operation(upload_id=upload.upload_id, payload=dict(operation='publish')).response()

    return upload.upload_id


@client.command(
    help='Upload files to nomad. The given path can be a single file or a directory. '
    'All .zip files in a directory will be uploaded.')
@click.argument('PATH', nargs=-1, required=True, type=click.Path(exists=True))
@click.option(
    '--name',
    help='Optional name for the upload of a single file. Will be ignored on directories.')
@click.option(
    '--offline', is_flag=True, default=False,
    help='Upload files "offline": files will not be uploaded, but processed were they are. '
    'Only works when run on the nomad host.')
@click.option(
    '--publish', is_flag=True, default=False,
    help='Automatically move upload out of the staging area after successful processing')
def upload(path, name: str, offline: bool, publish: bool):
    paths = path
    click.echo('uploading files from %s paths' % len(paths))
    for path in paths:
        click.echo('uploading %s' % path)
        if os.path.isfile(path):
            name = name if name is not None else os.path.basename(path)
            upload_file(path, name, offline, publish)

        elif os.path.isdir(path):
            for (dirpath, _, filenames) in os.walk(path):
                for filename in filenames:
                    if filename.endswith('.zip'):
                        file_path = os.path.abspath(os.path.join(dirpath, filename))
                        name = os.path.basename(file_path)
                        upload_file(file_path, name, offline, publish)

        else:
            click.echo('Unknown path type %s.' % path)



# Copyright 2018 Markus Scheidgen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an"AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tarfile

from .file_parser import FileParser


class TarParser(FileParser):
    def __init__(self, mainfile=None, logger=None):
        super().__init__(mainfile, logger, tarfile.open)
        self._names_map = None

    @property
    def names_map(self):
        if self._names_map is None:
            if self.mainfile_obj is not None:
                self._names_map = {f.lower(): f for f in self.mainfile_obj.getnames()}
        return self._names_map

    def parse(self, key):
        if self._results is None:
            self._results = dict()

        if self.mainfile_obj is None:
            return

        name = self.names_map.get(key, key)
        try:
            val = self.mainfile_obj.extractfile(name)
        except Exception:
            val = None

        self._results[key] = val



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import numpy as np

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class FleurParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*\-*\+*\d*'

        def get_forces(val):
            forces = re.findall(rf'FX\_TOT\=\s*({re_f}) FY\_TOT\=\s*({re_f}) FZ\_TOT\=\s*({re_f})', val)
            return np.array(forces, dtype=np.dtype(np.float64))

        super().__init__(
            specifications=dict(
                name='parsers/fleur', code_name='fleur',
                code_homepage='https://www.flapw.de/', domain='dft',
                mainfile_contents_re=r'This output is generated by fleur.'),
            units_mapping=dict(length=ureg.bohr, energy=ureg.hartree),
            program_version=r'This output is generated by (fleur[\w\.]+)',
            lattice_vectors=r'bravais matrices of real and reciprocal lattices\s*([\s\S]+?)\n *\n',
            atom_labels_atom_positions_scaled=rf'([A-Z][a-z]* +\d+ +\d+ +\d+ +\d+ +{re_f} +{re_f}\s*{re_f} +{re_f} +{re_f} +{re_f}[\s\S]+?)\-\-\-\-\-',
            energy_reference_fermi=(rf'\n *Fermi\s*\:\s*({re_f})', lambda x: [x]),
            energy_total=rf'\-\-\-\-\> total energy\= *({re_f})',
            energy_free=rf'\-\-\-\-\> free energy\= *({re_f})',
            atom_forces=(rf'TOTAL FORCES ON ATOMS \*+\s+([\s\S]+?)\n *\n', get_forces))



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
import numpy as np

from nomad.units import ureg
from nomad.parsing.file_parser import BasicParser


class DLPolyParser(BasicParser):
    def __init__(self):
        re_f = r'\-*\d+\.\d+E*\-*\+*\d*'

        def to_traj(val_in):
            labels = re.findall(r'([A-Z][a-z]*)[\+\-]* +\d+', val_in)
            traj = dict(atom_labels=labels)
            try:
                values = re.findall(rf'({re_f} +{re_f} +{re_f})', val_in)
                traj['lattice_vectors'] = np.array([v.split() for v in values[:3]], dtype=np.dtype(np.float64))
                values = np.array([v.split() for v in values[3:]], dtype=np.dtype(np.float64))
                values = np.reshape(values, (len(labels), len(values) // len(labels), 3))
                values = np.transpose(values, (1, 0, 2))
            except Exception:
                pass
            keys = ['atom_positions', 'atom_velocities', 'atom_forces']
            for n, value in enumerate(values):
                traj[keys[n]] = value
            return traj

        super().__init__(
            specifications=dict(
                name='parsers/dl-poly', code_name='DL_POLY',
                code_homepage='https://www.scd.stfc.ac.uk/Pages/DL_POLY.aspx',
                mainfile_contents_re=(r'\*\* DL_POLY \*\*')),
            units_mapping=dict(
                length=ureg.angstrom, energy=10 * ureg.joule / 6.02214076e+23,
                time=ureg.ps, mass=ureg.amu),
            auxilliary_files=r'(HISTORY)',
            program_version=r'\* +version\: +([\d\.]+.+?) +\*',
            lattice_vectors_atom_labels_atom_positions_atom_velocities_atom_forces=(
                r'step +\d+ +\d+ +\d+ +\d+ +.+([\s\S]+?)time',
                to_traj),
            energy_total=rf'\-+\s+\d+ +({re_f}) +(?:{re_f}\s+)+?{re_f}\n *\n')
